@@ -6,12 +6,11 @@ extern crate quote;
 extern crate syn;
 
 use proc_macro::TokenStream;
-use proc_macro2::Span;
 use proc_macro2::TokenStream as TokenStream2;
 use quote::quote;
+use syn::parse_macro_input;
 use syn::parse_quote;
 use syn::spanned::Spanned;
-use syn::parse_macro_input;
 
 // ---
 
@@ -42,9 +41,6 @@ fn clonepy_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2
         impl ClonePy for #name {
             fn clone_py(&self, py: Python) -> Self {
                 use self::#name::*;
-                let gil = pyo3::Python::acquire_gil();
-                let py = gil.python();
-
                 match self {
                     #(#variants,)*
                 }
@@ -71,18 +67,112 @@ fn clonepy_impl_struct(ast: &syn::DeriveInput, _en: &syn::DataStruct) -> TokenSt
 
 // ---
 
+// ---
+
+#[proc_macro_derive(EqPy)]
+pub fn eqpy_derive(input: TokenStream) -> TokenStream {
+    let ast = parse_macro_input!(input as syn::DeriveInput);
+    match &ast.data {
+        syn::Data::Enum(e) => TokenStream::from(eqpy_impl_enum(&ast, &e)),
+        syn::Data::Struct(s) => TokenStream::from(eqpy_impl_struct(&ast, &s)),
+        _ => panic!("#[derive(EqPy)] only supports enums or structs"),
+    }
+}
+
+fn eqpy_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2 {
+    let mut variants = Vec::new();
+
+    // Build eq_py for each variant
+    for variant in &en.variants {
+        let name = &variant.ident;
+        variants.push(quote! {
+            ( #name(l), #name(r) ) => (*l.bind(py).borrow()).eq_py(&*r.bind(py).borrow(), py)
+        });
+    }
+
+    // Build eq implementation
+    let name = &ast.ident;
+    let expanded = quote! {
+        #[automatically_derived]
+        #[allow(unused)]
+        impl EqPy for #name {
+            fn eq_py(&self, other: &Self, py: Python) -> bool {
+                use self::#name::*;
+                match (self, other) {
+                    #(#variants,)*
+                    _ => false
+                }
+            }
+        }
+    };
+
+    expanded
+}
+
+fn eqpy_impl_struct(ast: &syn::DeriveInput, en: &syn::DataStruct) -> TokenStream2 {
+    let mut expression: syn::Expr = parse_quote!(true);
+
+    // let fields: Vec<_> = match &en.fields {
+    //     syn::Fields::Named(n) => n.named.iter().map(|field| field.ident.as_ref().unwrap()).collect(),
+    //     _ => unreachable!(),
+    // };
+
+    if let syn::Fields::Named(n) = &en.fields {
+        for field in n.named.iter() {
+            let name = field.ident.as_ref().unwrap();
+            let condition = parse_quote!(self.#name.eq_py(&other.#name, py));
+            expression = syn::Expr::from(syn::ExprBinary {
+                attrs: Vec::new(),
+                left: Box::new(expression),
+                op: syn::BinOp::And(<syn::Token![&&]>::default()),
+                right: Box::new(condition),
+            })
+        }
+    } else {
+        unreachable!()
+    }
+
+    let name = &ast.ident;
+    let expanded = quote! {
+        #[automatically_derived]
+        impl EqPy for #name {
+            fn eq_py(&self, other: &Self, py: Python) -> bool {
+                #expression
+            }
+        }
+    };
+
+    expanded
+}
+
+// ---
+
 #[proc_macro_derive(PyWrapper, attributes(wraps))]
 pub fn pywrapper_derive(input: TokenStream) -> TokenStream {
     let ast = parse_macro_input!(input as syn::DeriveInput);
+
+    let base = {
+        let mut base = None;
+        for attr in &ast.attrs {
+            match &attr.meta {
+                syn::Meta::List(list)
+                    if list.path.get_ident().map(|i| i == "wraps").unwrap_or(false) =>
+                {
+                    base = Some(syn::parse2::<syn::Ident>(list.tokens.clone()).unwrap());
+                }
+                _ => (),
+            }
+        }
+        base.expect("failed to locate #[wraps(...)] attribute")
+    };
+
     let mut output = TokenStream2::new();
     if let syn::Data::Enum(e) = &ast.data {
-        // output.extend(clone_impl_enum(&ast, &e));
-        output.extend(topyobject_impl_enum(&ast, &e));
         output.extend(intopyobject_impl_enum(&ast, &e));
         output.extend(frompyobject_impl_enum(&ast, &e));
-        output.extend(aspyptr_impl_enum(&ast, &e));
         output.extend(intopy_impl_enum(&ast, &e));
-        // output.extend(pyobjectprotocol_impl_enum(&ast, &e))
+        output.extend(asbase_impl_enum(&ast, &e, &base));
+        output.extend(fromvariant_impl_enum(&ast, &e));
     } else {
         panic!("only supports enums");
     }
@@ -90,80 +180,52 @@ pub fn pywrapper_derive(input: TokenStream) -> TokenStream {
     TokenStream::from(output)
 }
 
-fn aspyptr_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2 {
-    let mut variants = Vec::new();
-
-    // Build clone for each variant
-    for variant in &en.variants {
-        let name = &variant.ident;
-        variants.push(quote!(#name(x) => x.as_ptr()));
-    }
-
-    // Build clone implementation
-    let name = &ast.ident;
-    let expanded = quote! {
-        #[automatically_derived]
-        impl pyo3::AsPyPointer for #name {
-            fn as_ptr(&self) -> *mut pyo3::ffi::PyObject {
-                use self::#name::*;
-
-                match self {
-                    #(#variants,)*
-                }
-            }
-        }
-    };
-
-    expanded
-}
-
-fn topyobject_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2 {
-    let mut variants = Vec::new();
-
-    // Build clone for each variant
-    for variant in &en.variants {
-        let name = &variant.ident;
-        variants.push(quote!(#name(x) => x.to_object(py)));
-    }
-
-    // Build clone implementation
-    let name = &ast.ident;
-    let expanded = quote! {
-        #[automatically_derived]
-        impl pyo3::ToPyObject for #name {
-            fn to_object(&self, py: Python) -> pyo3::PyObject {
-                use self::#name::*;
-                match self {
-                    #(#variants,)*
-                }
-            }
-        }
-    };
-
-    expanded
-}
-
 fn intopyobject_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2 {
     let mut variants = Vec::new();
 
+    // extract name of base struct
+    let meta = &ast
+        .attrs
+        .iter()
+        .find(|attr| attr.path().is_ident(&syn::Ident::new("wraps", attr.span())))
+        .expect("could not find #[wraps] attribute")
+        .meta;
+    let base: syn::Ident = match meta {
+        syn::Meta::List(l) => syn::parse2(l.tokens.clone()).unwrap(),
+        _ => panic!("#[wraps] argument must be a class ident"),
+    };
+
     // Build clone for each variant
     for variant in &en.variants {
         let name = &variant.ident;
-        variants.push(quote!(#name(x) => pyo3::IntoPy::into_py(x, py)));
+        variants.push(quote!(#name(x) => x.extract(py).map_err(PyErr::from)));
     }
 
     // Build clone implementation
     let name = &ast.ident;
     let expanded = quote! {
         #[automatically_derived]
-        impl pyo3::IntoPy<pyo3::PyObject> for #name {
-            fn into_py(self, py: Python) -> pyo3::PyObject {
+        impl<'py> pyo3::IntoPyObject<'py> for &#name {
+            type Error = PyErr;
+            type Target = #base;
+            type Output = Bound<'py, Self::Target>;
+            fn into_pyobject(self, py: Python<'py>) -> Result<Self::Output, Self::Error> {
                 use self::#name::*;
                 match self {
                     #(#variants,)*
                 }
             }
         }
+
+        #[automatically_derived]
+        impl<'py> pyo3::IntoPyObject<'py> for #name {
+            type Error = PyErr;
+            type Target = #base;
+            type Output = Bound<'py, Self::Target>;
+            fn into_pyobject(self, py: Python<'py>) -> Result<Self::Output, Self::Error> {
+                (&self).into_pyobject(py)
+            }
+        }
     };
 
     expanded
@@ -197,24 +259,18 @@ fn frompyobject_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenSt
             path.segments.iter().next().unwrap().ident.span(),
         );
 
-        variants.push(
-            quote!(#lit => ob.extract::<pyo3::Py<#path>>().map(#wrapped::#name))
-        );
+        variants.push(quote!(#lit => obj.extract::<pyo3::Py<#path>>().map(#wrapped::#name).map_err(PyErr::from)));
     }
 
-    let meta = ast
+    // extract name of base struct
+    let meta = &ast
         .attrs
         .iter()
-        .find(|attr| attr.path.is_ident(&syn::Ident::new("wraps", attr.span())))
+        .find(|attr| attr.path().is_ident(&syn::Ident::new("wraps", attr.span())))
         .expect("could not find #[wraps] attribute")
-        .parse_meta()
-        .expect("could not parse #[wraps] argument");
-
-    let base = match meta {
-        syn::Meta::List(l) => match l.nested.iter().next().unwrap() {
-            syn::NestedMeta::Meta(syn::Meta::Path(p)) => p.clone(),
-            _ => panic!("#[wraps] argument must be a class ident"),
-        },
+        .meta;
+    let base: syn::Ident = match meta {
+        syn::Meta::List(l) => syn::parse2(l.tokens.clone()).unwrap(),
         _ => panic!("#[wraps] argument must be a class ident"),
     };
 
@@ -229,17 +285,18 @@ fn frompyobject_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenSt
     );
     let expanded = quote! {
         #[automatically_derived]
-        impl<'source> pyo3::FromPyObject<'source> for #wrapped {
-            fn extract(ob: &'source pyo3::types::PyAny) -> pyo3::PyResult<Self> {
-                use pyo3::AsPyPointer;
+        impl<'source, 'py> pyo3::FromPyObject<'source, 'py> for #wrapped {
+            type Error = PyErr;
+            fn extract(obj: Borrowed<'source, 'py, PyAny>) -> pyo3::PyResult<Self> {
+                let qualname = obj.get_type().name()?;
+                let q = qualname.to_str()?;
 
-                let qualname = ob.get_type().name()?;
-                let ty = match qualname.rfind('.') {
-                    Some(idx) => &qualname[idx+1..],
-                    None => &qualname,
+                let ty = match q.rfind('.') {
+                    Some(idx) => &q[idx+1..],
+                    None => &q,
                 };
 
-                if ob.is_instance::<#base>()? {
+                if obj.is_instance_of::<#base>() {
                     match ty.as_ref() {
                         #(#variants,)*
                         _ => Err(pyo3::exceptions::PyTypeError::new_err(#err_sub))
@@ -247,7 +304,7 @@ fn frompyobject_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenSt
                 } else {
                     Err(pyo3::exceptions::PyTypeError::new_err(format!(
                         #err_ty,
-                        ob.get_type().name()?,
+                        obj.get_type().name()?,
                     )))
                 }
             }
@@ -264,15 +321,15 @@ fn intopy_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2
     for variant in &en.variants {
         let name = &variant.ident;
         variants.push(quote!(
-            #name(x) => (&x.as_ref(py).borrow()).clone_py(py).into_py(py)
+            #name(x) => (&x.bind(py).borrow()).clone_py(py).into_py(py)
         ));
     }
 
-    // Build clone implementation
+    // Build implementation
     let name = &ast.ident;
     let expanded = quote! {
         #[automatically_derived]
-        impl pyo3::IntoPy<fastobo::ast::#name> for &#name {
+        impl IntoPy<fastobo::ast::#name> for &#name {
             fn into_py(self, py: Python) -> fastobo::ast::#name {
                 use std::ops::Deref;
                 use self::#name::*;
@@ -286,55 +343,108 @@ fn intopy_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2
     expanded
 }
 
+fn asbase_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum, base: &syn::Ident) -> TokenStream2 {
+    let mut variants = Vec::new();
+
+    // Build IntoPy for each variant
+    for variant in &en.variants {
+        let name = &variant.ident;
+        variants.push(quote!(
+            #name(x) => x.bind(py).as_super()
+        ));
+    }
+
+    // Build implementation
+    let name = &ast.ident;
+    let expanded = quote! {
+        #[automatically_derived]
+        impl #name {
+            fn as_base<'a, 'py>(&'a self, py: Python<'py>) -> &'a Bound<'py, #base> {
+                use std::ops::Deref;
+                use self::#name::*;
+                match self {
+                    #(#variants,)*
+                }
+            }
+        }
+    };
+
+    expanded
+}
+
+fn fromvariant_impl_enum(ast: &syn::DeriveInput, en: &syn::DataEnum) -> TokenStream2 {
+    // extract name of enum wrapper
+    let wrapper = &ast.ident;
+
+    // Build From for each variant
+    let mut output = TokenStream2::new();
+    for variant in &en.variants {
+        let name = &variant.ident;
+        let ty = if let syn::Fields::Unnamed(f) = &variant.fields {
+            if let syn::Type::Path(p) = &f.unnamed.first().unwrap().ty {
+                if let syn::PathArguments::AngleBracketed(b) =
+                    &p.path.segments.first().unwrap().arguments
+                {
+                    &b.args.first().unwrap()
+                } else {
+                    panic!("only supports bracketed paths")
+                }
+            } else {
+                panic!("only supports generic field types")
+            }
+        } else {
+            panic!("only supports unnamed fields");
+        };
+        output.extend(quote!(
+            #[automatically_derived]
+            impl From<Py<#ty>> for #wrapper {
+                fn from(x: Py<#ty>) -> Self {
+                    #wrapper::#name(x)
+                }
+            }
+            #[automatically_derived]
+            impl<'py> From<Bound<'py, #ty>> for #wrapper {
+                fn from(x: Bound<'py, #ty>) -> Self {
+                    #wrapper::#name(x.unbind())
+                }
+            }
+        ));
+    }
+
+    output
+}
+
 // ---
 
 #[proc_macro_attribute]
-pub fn listlike(attr: TokenStream, input: TokenStream) -> TokenStream {
-    // extract proc-macro arguments
-    let meta = parse_macro_input!(attr as syn::AttributeArgs);
-    let field: syn::Ident = if let syn::Lit::Str(ref s) = meta
-        .iter()
-        .filter_map(|m| match m {
-            syn::NestedMeta::Meta(m) => Some(m),
-            _ => None,
-        })
-        .filter_map(|m| match m {
-            syn::Meta::NameValue(nv) => Some(nv),
-            _ => None
-        })
-        .find(|nv| nv.path.get_ident() == Some(&syn::Ident::new("field", Span::call_site())))
-        .expect("#[pylist] requires a `field` argument")
-        .lit
-    {
-        s.parse().expect("`field` argument of #[pylist] is not a valid identifier")
-    } else {
-        panic!("`field` argument of #[pylist] must be a string");
-    };
-    let ty: syn::Type = if let syn::Lit::Str(ref s) = meta
-        .iter()
-        .filter_map(|m| match m {
-            syn::NestedMeta::Meta(m) => Some(m),
-            _ => None,
-        })
-        .filter_map(|m| match m {
-            syn::Meta::NameValue(nv) => Some(nv),
-            _ => None
-        })
-        .find(|nv| nv.path.get_ident() == Some(&syn::Ident::new("type", Span::call_site())))
-        .expect("#[pylist] requires a `type` argument")
-        .lit
-    {
-        s.parse().expect("`type` argument of #[pylist] is not a valid type")
-    } else {
-        panic!("`type` argument of #[pylist] must be a string");
-    };
+pub fn listlike(args: TokenStream, input: TokenStream) -> TokenStream {
+    let mut field: Option<syn::Ident> = None;
+    let mut ty: Option<syn::Type> = None;
+    let listlike_parser = syn::meta::parser(|meta| {
+        if meta.path.is_ident("field") {
+            let name: syn::LitStr = meta.value()?.parse()?;
+            field = Some(syn::parse_str(&name.value())?);
+            Ok(())
+        } else if meta.path.is_ident("type") {
+            let name: syn::LitStr = meta.value()?.parse()?;
+            ty = Some(syn::parse_str(&name.value())?);
+            Ok(())
+        } else {
+            Err(meta.error("unsupported listlike property"))
+        }
+    });
+    parse_macro_input!(args with listlike_parser);
 
     // add additional methods to the impl block
     let ast = parse_macro_input!(input as syn::ItemImpl);
-    TokenStream::from(listlike_impl_methods(&field, &ty, ast))
+    TokenStream::from(listlike_impl_methods(&field.unwrap(), &ty.unwrap(), ast))
 }
 
-fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemImpl) -> TokenStream2 {
+fn listlike_impl_methods(
+    field: &syn::Ident,
+    ty: &syn::Type,
+    mut imp: syn::ItemImpl,
+) -> TokenStream2 {
     imp.items.push(parse_quote! {
         /// Append object to the end of the list.
         ///
@@ -343,8 +453,8 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
         ///         this container (see type-level documentation for the
         ///         required type).
         #[pyo3(text_signature = "(self, object)")]
-        fn append(&mut self, object: &PyAny) -> PyResult<()> {
-            let item = <#ty as pyo3::prelude::FromPyObject>::extract(object)?;
+        fn append<'py>(&mut self, object: &Bound<'py, PyAny>) -> PyResult<()> {
+            let item = object.extract()?;
             self.#field.push(item);
             Ok(())
         }
@@ -360,9 +470,10 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
         /// Return a shallow copy of the list.
         #[pyo3(text_signature = "(self)")]
         fn copy(&self) -> PyResult<Py<Self>> {
-            let gil = Python::acquire_gil();
-            let copy = self.clone_py(gil.python());
-            Py::new(gil.python(), copy)
+            Python::with_gil(|py| {
+                let copy = self.clone_py(py);
+                Py::new(py, copy)
+            })
         }
     });
     imp.items.push(parse_quote! {
@@ -373,9 +484,10 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
         ///         this container (see type-level documentation for the
         ///         required type).
         #[pyo3(text_signature = "(self, value)")]
-        fn count(&mut self, value: &PyAny) -> PyResult<usize> {
-            let item = <#ty as pyo3::prelude::FromPyObject>::extract(value)?;
-            Ok(self.#field.iter().filter(|&x| *x == item).count())
+        fn count<'py>(&mut self, value: &Bound<'py, PyAny>) -> PyResult<usize> {
+            let py = value.py();
+            let item = value.extract()?;
+            Ok(self.#field.iter().filter(|&x| x.eq_py(&item, py)).count())
         }
     });
     // |  extend($self, iterable, /)
@@ -392,8 +504,8 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
         /// If `index` is greater than the number of elements in the list,
         /// `object` will be added at the end of the list.
         #[pyo3(text_signature = "(self, index, object)")]
-        fn insert(&mut self, mut index: isize, object: &PyAny) -> PyResult<()> {
-            let item = <#ty as pyo3::prelude::FromPyObject>::extract(object)?;
+        fn insert<'py>(&mut self, mut index: isize, object: &Bound<'py, PyAny>) -> PyResult<()> {
+            let item = object.extract()?;
             if index >= self.#field.len() as isize {
                 self.#field.push(item);
             } else {
@@ -410,8 +522,7 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
         ///
         /// Raises:
         ///     IndexError: when list is empty or index is out of range.
-        #[args(index="-1")]
-        #[pyo3(text_signature = "(self, index=-1)")]
+        #[pyo3(text_signature = "(self, index=-1)", signature=(index=-1))]
         fn pop(&mut self, mut index: isize) -> PyResult<#ty> {
             // Wrap once to allow negative indexing
             if index < 0 {
@@ -441,7 +552,6 @@ fn listlike_impl_methods(field: &syn::Ident, ty: &syn::Type, mut imp: syn::ItemI
     quote!(#imp)
 }
 
-
 // ---
 
 #[proc_macro_derive(FinalClass, attributes(base))]
@@ -454,39 +564,39 @@ pub fn finalclass_derive(input: TokenStream) -> TokenStream {
 }
 
 fn finalclass_impl_struct(ast: &syn::DeriveInput, _st: &syn::DataStruct) -> TokenStream2 {
-    // Get the name of the wrapped struct.
-    let name = &ast.ident;
-
-    // Get the name of the base class.
-    let meta = ast
+    // Get the `base` attribute.
+    let meta = &ast
         .attrs
         .iter()
-        .find(|attr| attr.path.is_ident(&syn::Ident::new("base", attr.span())))
+        .find(|attr| attr.path().is_ident(&syn::Ident::new("base", attr.span())))
         .expect("could not find #[base] attribute")
-        .parse_meta()
-        .expect("could not parse #[base] argument");
-    let base = match meta {
-        syn::Meta::List(l) => match l.nested.iter().next().unwrap() {
-            syn::NestedMeta::Meta(syn::Meta::Path(p)) => p.clone(),
-            _ => panic!("#[base] argument must be a class ident"),
-        },
+        .meta;
+    let base: syn::Type = match meta {
+        syn::Meta::List(l) => syn::parse2(l.tokens.clone()).unwrap(),
         _ => panic!("#[base] argument must be a class ident"),
     };
 
+    // Get the name of the wrapped struct.
+    let name = &ast.ident;
+
     // derive an implementation of PyClassInitializer using simply the
     // default value of the base class as the initializer value
     quote! {
         impl FinalClass for #name {}
         impl Into<pyo3::pyclass_init::PyClassInitializer<#name>> for #name {
             fn into(self) ->  pyo3::pyclass_init::PyClassInitializer<Self> {
-                <#base as AbstractClass>::initializer()
-                    .add_subclass(self)
+                Python::attach(|py| self.into_py(py))
+            }
+        }
+
+        impl IntoPy<pyo3::pyclass_init::PyClassInitializer<#name>> for #name {
+            fn into_py(self, py: Python) ->  pyo3::pyclass_init::PyClassInitializer<Self> {
+                #base::initializer(py).add_subclass(self)
             }
         }
     }
 }
 
-
 // ---
 
 #[proc_macro_derive(AbstractClass, attributes(base))]
@@ -499,32 +609,27 @@ pub fn abstractclass_derive(input: TokenStream) -> TokenStream {
 }
 
 fn abstractclass_impl_struct(ast: &syn::DeriveInput, _st: &syn::DataStruct) -> TokenStream2 {
-    // Get the name of the wrapped struct.
-    let name = &ast.ident;
-
-    // Get the name of the base class.
-    let meta = ast
+    // Get the `base` attribute.c
+    let meta = &ast
         .attrs
         .iter()
-        .find(|attr| attr.path.is_ident(&syn::Ident::new("base", attr.span())))
+        .find(|attr| attr.path().is_ident(&syn::Ident::new("base", attr.span())))
         .expect("could not find #[base] attribute")
-        .parse_meta()
-        .expect("could not parse #[base] argument");
-    let base = match meta {
-        syn::Meta::List(l) => match l.nested.iter().next().unwrap() {
-            syn::NestedMeta::Meta(syn::Meta::Path(p)) => p.clone(),
-            _ => panic!("#[base] argument must be a class ident"),
-        },
+        .meta;
+    let base: syn::Type = match meta {
+        syn::Meta::List(l) => syn::parse2(l.tokens.clone()).unwrap(),
         _ => panic!("#[base] argument must be a class ident"),
     };
 
+    // Get the name of the wrapped struct.
+    let name = &ast.ident;
+
     // derive an implementation of PyClassInitializer using simply the
     // default value of the base class as the initializer value
     quote! {
         impl AbstractClass for #name {
-            fn initializer() -> pyo3::pyclass_init::PyClassInitializer<Self> {
-                <#base as AbstractClass>::initializer()
-                    .add_subclass(Self {})
+            fn initializer(py: Python) -> pyo3::pyclass_init::PyClassInitializer<Self> {
+                #base::initializer(py).add_subclass(Self::default())
             }
         }
 
