@@ -6,31 +6,28 @@ use std::io::BufReader;
 use std::io::Error as IoError;
 use std::io::Read;
 use std::iter::Iterator;
-use std::path::Path;
-use std::path::PathBuf;
 use std::num::NonZeroUsize;
 use std::ops::Deref;
 use std::ops::DerefMut;
+use std::path::Path;
+use std::path::PathBuf;
 
 use pyo3::exceptions::PyValueError;
 use pyo3::prelude::*;
-use pyo3::types::PyString;
 use pyo3::types::PyBytes;
-use pyo3::PyIterProtocol;
-use pyo3::PyObjectProtocol;
-use pyo3::AsPyPointer;
-use pyo3::PyGCProtocol;
+use pyo3::types::PyString;
 
 use fastobo::parser::Parser;
-use fastobo::parser::ThreadedParser;
 use fastobo::parser::SequentialParser;
+use fastobo::parser::ThreadedParser;
 
 use crate::error::Error;
-use crate::py::header::frame::HeaderFrame;
 use crate::py::doc::EntityFrame;
+use crate::py::header::frame::HeaderFrame;
 use crate::pyfile::PyFileGILRead;
 use crate::transmute_file_error;
 use crate::utils::ClonePy;
+use crate::utils::IntoPy;
 
 // ---------------------------------------------------------------------------
 
@@ -41,14 +38,15 @@ pub enum Handle {
 }
 
 impl Handle {
-    fn handle(&self) -> PyObject {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
+    fn handle<'py>(&self, py: Python<'py>) -> PyResult<Bound<'py, PyAny>> {
         match self {
-            Handle::FsFile(_, path) =>  path.display().to_string().to_object(py),
-            Handle::PyFile(f) => {
-                f.file().lock().unwrap().to_object(py)
-            }
+            Handle::FsFile(_, path) => path
+                .display()
+                .to_string()
+                .into_pyobject(py)
+                .map(|b| b.into_any())
+                .map_err(PyErr::from),
+            Handle::PyFile(f) => Ok(f.file().lock().unwrap().bind(py).clone()),
         }
     }
 }
@@ -83,13 +81,15 @@ impl<B: BufRead> InternalParser<B> {
         match n {
             0 => Ok(InternalParser::Threaded(ThreadedParser::new(stream))),
             1 => Ok(InternalParser::Sequential(SequentialParser::new(stream))),
-            n if n < 0 => {
-                Err(PyValueError::new_err("threads count must be positive or null"))
-            },
+            n if n < 0 => Err(PyValueError::new_err(
+                "threads count must be positive or null",
+            )),
             n => {
                 let t = std::num::NonZeroUsize::new(n as usize).unwrap();
-                Ok(InternalParser::Threaded(ThreadedParser::with_threads(stream, t)))
-            },
+                Ok(InternalParser::Threaded(ThreadedParser::with_threads(
+                    stream, t,
+                )))
+            }
         }
     }
 
@@ -182,19 +182,15 @@ pub struct FrameReader {
 
 impl FrameReader {
     fn new(handle: BufReader<Handle>, ordered: bool, threads: i16) -> PyResult<Self> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
         let mut inner = InternalParser::with_thread_count(handle, threads)?;
         inner.ordered(ordered);
         let frame = inner
             .next()
             .unwrap()
             .map_err(Error::from)?
-            .into_header_frame()
+            .into_header()
             .unwrap();
-        let header = Py::new(py, frame.into_py(py))?;
-
+        let header = Python::with_gil(|py| Py::new(py, frame.into_py(py)))?;
         Ok(Self { inner, header })
     }
 
@@ -206,7 +202,12 @@ impl FrameReader {
         }
     }
 
-    pub fn from_handle(obj: &PyAny, ordered: bool, threads: i16) -> PyResult<Self> {
+    pub fn from_handle<'py>(
+        obj: &Bound<'py, PyAny>,
+        ordered: bool,
+        threads: i16,
+    ) -> PyResult<Self> {
+        let py = obj.py();
         match PyFileGILRead::from_ref(obj).map(Handle::PyFile) {
             Ok(inner) => Self::new(BufReader::new(inner), ordered, threads),
             Err(e) => Err(e),
@@ -216,44 +217,34 @@ impl FrameReader {
 
 #[pymethods]
 impl FrameReader {
-    fn header<'py>(&self, py: Python<'py>) -> Py<HeaderFrame> {
-        self.header.clone_py(py)
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        let py = slf.py();
+        let fmt = PyString::intern(py, "fastobo.iter({!r})");
+        fmt.call_method1("format", (slf.inner.as_ref().get_ref().handle(py)?,))
     }
-}
 
-#[pyproto]
-impl PyObjectProtocol for FrameReader {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let fmt = PyString::new(py, "fastobo.iter({!r})").to_object(py);
-        fmt.call_method1(py, "format", (&self.inner.as_ref().get_ref().handle(),))
-    }
-}
-
-#[pyproto]
-impl PyIterProtocol for FrameReader {
-    fn __iter__(slf: PyRefMut<'p, Self>) -> PyResult<PyRefMut<'p, Self>> {
+    fn __iter__(slf: PyRefMut<Self>) -> PyResult<PyRefMut<Self>> {
         Ok(slf)
     }
 
-    fn __next__(mut slf: PyRefMut<'p, Self>) -> PyResult<Option<EntityFrame>> {
+    fn __next__(mut slf: PyRefMut<'_, Self>) -> PyResult<Option<EntityFrame>> {
         match slf.deref_mut().inner.next() {
             None => Ok(None),
             Some(Ok(frame)) => {
-                let gil = Python::acquire_gil();
-                let entity = frame.into_entity_frame().unwrap();
-                Ok(Some(entity.into_py(gil.python())))
-            },
-            Some(Err(e)) => {
-                let gil = Python::acquire_gil();
-                let py = gil.python();
+                let entity = frame.into_entity().unwrap();
+                Ok(Some(Python::with_gil(|py| entity.into_py(py))))
+            }
+            Some(Err(e)) => Python::with_gil(|py| {
                 if PyErr::occurred(py) {
                     Err(PyErr::fetch(py))
                 } else {
                     Err(Error::from(e).into())
                 }
-            }
+            }),
         }
     }
+
+    fn header<'py>(&self, py: Python<'py>) -> Py<HeaderFrame> {
+        self.header.clone_py(py)
+    }
 }
