@@ -8,49 +8,71 @@ macro_rules! impl_hash {
     });
 }
 
-macro_rules! impl_richmp {
+macro_rules! impl_richcmp {
     ($self:ident, $other:ident, $op:ident, $(self . $attr:ident)&&*) => ({
+        use pyo3::types::PyBool;
         match $op {
             $crate::pyo3::class::basic::CompareOp::Eq => {
-                if let Ok(ref clause) = $other.extract::<Py<Self>>() {
-                    let clause = clause.as_ref($other.py()).borrow();
-                    Ok(($($self.$attr == clause.$attr)&&*).to_object($other.py()))
+                let py = $other.py();
+                let val = if let Ok(ref clause) = $other.cast::<Self>() {
+                    let c = clause.borrow();
+                    $($self.$attr == c.$attr)&&*
                 } else {
-                    Ok(false.to_object($other.py()))
-                }
-            }
-            $crate::pyo3::class::basic::CompareOp::Ne => {
-                if let Ok(ref clause) = $other.extract::<Py<Self>>() {
-                    let clause = clause.as_ref($other.py()).borrow();
-                    Ok(($($self.$attr != clause.$attr)||*).to_object($other.py()))
-                } else {
-                    Ok(true.to_object($other.py()))
-                }
+                    false
+                };
+                Ok(PyBool::new(py, val).to_owned().into_any().unbind())
             }
             _ => Ok($other.py().NotImplemented())
         }
     });
 }
 
-macro_rules! impl_repr {
+macro_rules! impl_richcmp_py {
+    ($self:ident, $other:ident, $op:ident, $(self . $attr:ident)&&*) => ({
+        use pyo3::types::PyBool;
+        match $op {
+            $crate::pyo3::class::basic::CompareOp::Eq => {
+                let py = $other.py();
+                let val = if let Ok(ref clause) = $other.cast::<Self>() {
+                    let c = clause.borrow();
+                    $($self.$attr.eq_py(&c.$attr, py))&&*
+                } else {
+                    false
+                };
+                Ok(PyBool::new(py, val).to_owned().into_any().unbind())
+            }
+            _ => Ok($other.py().NotImplemented())
+        }
+    });
+}
+
+macro_rules! impl_repr_py {
     ($self:ident, $cls:ident($($field:expr),*)) => ({
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
+        let py = $self.py();
         let args = &[
-            $($field.to_object(py).as_ref(py).repr()?.to_str()?,)*
+            $((&$field).into_pyobject(py)?.as_any().repr()?.to_str()?,)*
         ].join(", ");
-
-        Ok(PyString::new(py, &format!("{}({})", stringify!($cls), args)).to_object(py))
+        Ok(PyString::new(py, &format!("{}({})", stringify!($cls), args)).into_pyobject(py)?.into_any())
     })
 }
 
+macro_rules! impl_str_py {
+    ($self:ident, $cls:ident) => {{
+        let py = $self.py();
+        let line: fastobo::ast::Line<fastobo::ast::$cls> = $cls::from($self).into_py(py);
+        let mut s = line.to_string();
+        if s.ends_with('\n') {
+            s.pop();
+        }
+        s
+    }};
+}
+
 macro_rules! register {
     ($py:ident, $m:ident, $cls:ident, $module:expr, $metacls:ident) => {
         $py.import($module)?
             .getattr(stringify!($metacls))?
-            .to_object($py)
-            .call_method1($py, "register", ($m.getattr(stringify!($cls))?,))?;
+            .call_method1("register", ($m.getattr(stringify!($cls))?,))?;
     };
 }
 
@@ -64,12 +86,12 @@ macro_rules! add_submodule {
         module.add("__package__", $sup.getattr("__package__")?)?;
 
         // add the submodule to the supermodule
-        $sup.add(stringify!($sub), module)?;
+        $sup.add(stringify!($sub), &module)?;
 
         // add the submodule to the `sys.modules` index
         $py.import("sys")?
             .getattr("modules")?
-            .cast_as::<pyo3::types::PyDict>()?
-            .set_item(concat!("fastobo.", stringify!($sub)), module)?;
+            .downcast::<pyo3::types::PyDict>()?
+            .set_item(concat!("fastobo.", stringify!($sub)), &module)?;
     }};
 }
