@@ -12,22 +12,20 @@ use pyo3::prelude::*;
 use pyo3::types::PyAny;
 use pyo3::types::PyList;
 use pyo3::types::PyString;
-use pyo3::PyGCProtocol;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use fastobo::ast as obo;
 
 use crate::error::Error;
-use crate::utils::ClonePy;
+use crate::py::qual::QualifierList;
 use crate::utils::AbstractClass;
+use crate::utils::ClonePy;
 
-use super::header::frame::HeaderFrame;
+// use super::header::frame::HeaderFrame;
+use super::id::BaseIdent;
 use super::id::Ident;
-use super::term::frame::TermFrame;
-use super::typedef::frame::TypedefFrame;
+// use super::term::frame::TermFrame;
+// use super::typedef::frame::TypedefFrame;
 
 // --- Module export ---------------------------------------------------------
 
@@ -42,7 +40,7 @@ use super::typedef::frame::TypedefFrame;
 ///
 #[pymodule]
 #[pyo3(name = "abc")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::AbstractFrame>()?;
     m.add_class::<self::AbstractEntityFrame>()?;
     m.add_class::<self::AbstractClause>()?;
@@ -64,7 +62,7 @@ pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
 pub struct AbstractFrame {}
 
 impl AbstractClass for AbstractFrame {
-    fn initializer() -> PyClassInitializer<Self> {
+    fn initializer(py: Python) -> PyClassInitializer<Self> {
         PyClassInitializer::from(Self {})
     }
 }
@@ -84,8 +82,10 @@ pub struct AbstractEntityFrame {}
 impl AbstractEntityFrame {
     /// `~fastobo.id.Ident`: the identifier of the described entity.
     #[getter]
-    pub fn get_id(&self) -> PyResult<Ident> {
-        Err(PyNotImplementedError::new_err("AbstractEntityFrame.raw_tag"))
+    pub fn get_id(&self) -> PyResult<BaseIdent> {
+        Err(PyNotImplementedError::new_err(
+            "AbstractEntityFrame.raw_tag",
+        ))
     }
 }
 
@@ -109,21 +109,21 @@ impl AbstractEntityFrame {
 ///      'xref': ['value-type:xsd\\:string "The allowed value-type for this CV term."']}
 ///
 #[pyclass(subclass, module = "fastobo.abc")]
-#[derive(Default)]
+#[derive(Debug, Default)]
 pub struct AbstractClause {}
 
 impl AbstractClass for AbstractClause {
-    fn initializer() -> PyClassInitializer<Self> {
+    fn initializer(py: Python) -> PyClassInitializer<Self> {
         PyClassInitializer::from(Self {})
     }
 }
 
 #[pymethods]
 impl AbstractClause {
-    /// Get the raw tag of the header clause.
+    /// Get the raw tag of the clause.
     ///
     /// Returns:
-    ///     `str`: the header clause value as it was extracted from the OBO
+    ///     `str`: the clause tag as it was extracted from the OBO
     ///     header, stripped from trailing qualifiers and comment.
     ///
     /// Example:
@@ -136,10 +136,10 @@ impl AbstractClause {
         Err(PyNotImplementedError::new_err("BaseHeaderClause.raw_tag"))
     }
 
-    /// Get the raw value of the header clause.
+    /// Get the raw value of the clause.
     ///
     /// Returns:
-    ///     `str`: the header clause value as it was extracted from the OBO
+    ///     `str`: the clause value as it was extracted from the OBO
     ///     header, stripped from trailing qualifiers and comment.
     ///
     /// Example:
@@ -156,6 +156,34 @@ impl AbstractClause {
 
 /// An abstract entity clause.
 #[pyclass(subclass, extends=AbstractClause, module="fastobo.abc")]
-#[derive(Default, AbstractClass)]
-#[base(AbstractClause)]
-pub struct AbstractEntityClause {}
+#[derive(Debug)]
+pub struct AbstractEntityClause {
+    #[pyo3(set)]
+    pub(crate) qualifiers: Option<Py<QualifierList>>,
+    #[pyo3(set)]
+    pub(crate) comment: Option<String>,
+}
+
+impl AbstractClass for AbstractEntityClause {
+    fn initializer(py: Python) -> PyClassInitializer<Self> {
+        AbstractClause::initializer(py).add_subclass(Self {
+            qualifiers: None,
+            comment: None,
+        })
+    }
+}
+
+#[pymethods]
+impl AbstractEntityClause {
+    /// `QualifierList` or `None`: the end-of-line qualifiers for this clause.
+    #[getter]
+    pub fn get_qualifiers(&self) -> &Option<Py<QualifierList>> {
+        &self.qualifiers
+    }
+
+    /// `str` or `None`: the end-of-line comment for this clause.
+    #[getter]
+    pub fn get_comment(&self) -> &Option<String> {
+        &self.comment
+    }
+}
