@@ -11,13 +11,9 @@ use pyo3::exceptions::PyIndexError;
 use pyo3::gc::PyTraverseError;
 use pyo3::prelude::*;
 use pyo3::types::PyAny;
+use pyo3::types::PyIterator;
 use pyo3::types::PyList;
 use pyo3::types::PyString;
-use pyo3::types::PyIterator;
-use pyo3::PyGCProtocol;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use fastobo::ast as obo;
@@ -25,18 +21,22 @@ use fastobo::visit::VisitMut;
 
 use crate::error::Error;
 use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::IntoPy;
 
+use super::abc::AbstractClause;
+use super::abc::AbstractEntityFrame;
 use super::abc::AbstractFrame;
 use super::header::frame::HeaderFrame;
+use super::instance::frame::InstanceFrame;
 use super::term::frame::TermFrame;
 use super::typedef::frame::TypedefFrame;
-use super::instance::frame::InstanceFrame;
 
 // --- Module export ---------------------------------------------------------
 
 #[pymodule]
 #[pyo3(name = "doc")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::OboDoc>()?;
     m.add("__name__", "fastobo.doc")?;
     Ok(())
@@ -44,8 +44,8 @@ pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
 
 // --- Conversion Wrapper ----------------------------------------------------
 
-#[derive(ClonePy, Debug, PartialEq, PyWrapper)]
-#[wraps(AbstractFrame)]
+#[derive(ClonePy, Debug, PyWrapper, EqPy)]
+#[wraps(AbstractEntityFrame)]
 pub enum EntityFrame {
     Term(Py<TermFrame>),
     Typedef(Py<TypedefFrame>),
@@ -63,7 +63,7 @@ impl IntoPy<EntityFrame> for fastobo::ast::EntityFrame {
             }
             fastobo::ast::EntityFrame::Instance(frame) => {
                 Py::new(py, frame.into_py(py)).map(EntityFrame::Instance)
-            },
+            }
         }
         .expect("could not allocate on Python heap")
     }
@@ -93,7 +93,7 @@ impl IntoPy<fastobo::ast::EntityFrame> for EntityFrame {
 ///         frames, either `TermFrame`, `TypedefFrame` or `InstanceFrame`.
 ///
 #[pyclass(module = "fastobo.doc")]
-#[derive(Debug)]
+#[derive(Debug, EqPy)]
 pub struct OboDoc {
     #[pyo3(get, set)]
     /// `~fastobo.header.HeaderFrame`: the header containing ontology metadata.
@@ -108,10 +108,7 @@ impl OboDoc {
     }
 
     pub fn with_entities(header: Py<HeaderFrame>, entities: Vec<EntityFrame>) -> Self {
-        Self {
-            header,
-            entities
-        }
+        Self { header, entities }
     }
 }
 
@@ -126,9 +123,7 @@ impl ClonePy for OboDoc {
 
 impl Display for OboDoc {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let doc: fastobo::ast::OboDoc = self.clone_py(py).into_py(py);
+        let doc: fastobo::ast::OboDoc = Python::with_gil(|py| self.clone_py(py).into_py(py));
         doc.fmt(f)
     }
 }
@@ -142,8 +137,7 @@ impl IntoPy<OboDoc> for fastobo::ast::OboDoc {
             .map(|frame| <fastobo::ast::EntityFrame as IntoPy<EntityFrame>>::into_py(frame, py))
             .collect();
 
-        let header = Py::new(py,h)
-            .expect("could not move header to Python heap");
+        let header = Py::new(py, h).expect("could not move header to Python heap");
 
         OboDoc { header, entities }
     }
@@ -151,10 +145,12 @@ impl IntoPy<OboDoc> for fastobo::ast::OboDoc {
 
 impl IntoPy<fastobo::ast::OboDoc> for OboDoc {
     fn into_py(self, py: Python) -> fastobo::ast::OboDoc {
-        let header: HeaderFrame = self.header.as_ref(py).borrow().clone_py(py);
+        let header: HeaderFrame = self.header.bind(py).borrow().clone_py(py);
         self.entities
             .iter()
-            .map(|frame| <EntityFrame as IntoPy<fastobo::ast::EntityFrame>>::into_py(frame.clone_py(py), py))
+            .map(|frame| {
+                <EntityFrame as IntoPy<fastobo::ast::EntityFrame>>::into_py(frame.clone_py(py), py)
+            })
             .collect::<fastobo::ast::OboDoc>()
             .and_header(header.into_py(py))
     }
@@ -164,23 +160,46 @@ impl IntoPy<fastobo::ast::OboDoc> for OboDoc {
 #[pymethods]
 impl OboDoc {
     #[new]
-    fn __init__(header: Option<&HeaderFrame>, entities: Option<&PyAny>) -> PyResult<Self> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
-        // extract header
-        let header = header
-            .map(|h| h.clone_py(py))
-            .unwrap_or_else(HeaderFrame::empty);
-        // create doc and extract entities
-        let mut doc = OboDoc::new(Py::from(PyCell::new(py, header)?));
-        if let Some(any) = entities {
-            for res in PyIterator::from_object(py, &any.to_object(py))? {
-                doc.entities.push(EntityFrame::extract(res?)?);
+    #[pyo3(signature = (header = None, entities = None))]
+    fn __init__<'py>(
+        header: Option<&HeaderFrame>,
+        entities: Option<&Bound<'py, PyAny>>,
+    ) -> PyResult<Self> {
+        Python::with_gil(|py| {
+            // extract header
+            let header = header
+                .map(|h| h.clone_py(py))
+                .unwrap_or_else(HeaderFrame::empty);
+            // create doc and extract entities
+            let mut doc = OboDoc::new(Py::new(py, header)?);
+            if let Some(any) = entities {
+                for res in PyIterator::from_object(&any)? {
+                    doc.entities.push(EntityFrame::extract(res?.as_borrowed())?);
+                }
             }
-        }
+            Ok(doc)
+        })
+    }
 
-        Ok(doc)
+    fn __str__(&self) -> PyResult<String> {
+        Ok(self.to_string())
+    }
+
+    fn __len__(&self) -> PyResult<usize> {
+        Ok(self.entities.len())
+    }
+
+    fn __getitem__<'py>(
+        &self,
+        py: Python<'py>,
+        index: isize,
+    ) -> PyResult<Bound<'py, AbstractEntityFrame>> {
+        if index < self.entities.len() as isize {
+            let item = &self.entities[index as usize];
+            item.into_pyobject(py)
+        } else {
+            Err(PyIndexError::new_err("list index out of range"))
+        }
     }
 
     #[getter]
@@ -221,11 +240,11 @@ impl OboDoc {
     ///
     #[pyo3(text_signature = "(self, /)")]
     fn compact_ids(&self) -> PyResult<Self> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let mut doc: obo::OboDoc = self.clone_py(py).into_py(py);
-        py.allow_threads(|| fastobo::visit::IdCompactor::new().visit_doc(&mut doc));
-        Ok(doc.into_py(py))
+        Python::with_gil(|py| {
+            let mut doc: obo::OboDoc = self.clone_py(py).into_py(py);
+            py.allow_threads(|| fastobo::visit::IdCompactor::new().visit_doc(&mut doc));
+            Ok(doc.into_py(py))
+        })
     }
 
     /// Create a semantically equivalent OBO document with IRI identifiers.
@@ -258,35 +277,10 @@ impl OboDoc {
     ///
     #[pyo3(text_signature = "(self, /)")]
     fn decompact_ids(&self) -> PyResult<Self> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let mut doc: obo::OboDoc = self.clone_py(py).into_py(py);
-        py.allow_threads(|| fastobo::visit::IdDecompactor::new().visit_doc(&mut doc));
-        Ok(doc.into_py(py))
-    }
-}
-
-#[pyproto]
-impl PyObjectProtocol for OboDoc {
-    fn __str__(&self) -> PyResult<String> {
-        Ok(self.to_string())
-    }
-}
-
-#[pyproto]
-impl PySequenceProtocol for OboDoc {
-    fn __len__(&self) -> PyResult<usize> {
-        Ok(self.entities.len())
-    }
-
-    fn __getitem__(&self, index: isize) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        if index < self.entities.len() as isize {
-            let item = &self.entities[index as usize];
-            Ok(item.to_object(py))
-        } else {
-            Err(PyIndexError::new_err("list index out of range"))
-        }
+        Python::with_gil(|py| {
+            let mut doc: obo::OboDoc = self.clone_py(py).into_py(py);
+            py.allow_threads(|| fastobo::visit::IdDecompactor::new().visit_doc(&mut doc));
+            Ok(doc.into_py(py))
+        })
     }
 }
