@@ -10,22 +10,19 @@ use pyo3::types::PyAny;
 use pyo3::types::PyIterator;
 use pyo3::types::PyList;
 use pyo3::types::PyString;
-use pyo3::AsPyPointer;
-use pyo3::PyGCProtocol;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use super::super::abc::AbstractFrame;
 use super::clause::BaseHeaderClause;
 use super::clause::HeaderClause;
-use crate::utils::ClonePy;
-use crate::utils::FinalClass;
 use crate::utils::AbstractClass;
+use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::FinalClass;
+use crate::utils::IntoPy;
 
 #[pyclass(extends=AbstractFrame, module="fastobo.header")]
-#[derive(Debug, FinalClass)]
+#[derive(Debug, FinalClass, EqPy)]
 #[base(AbstractFrame)]
 pub struct HeaderFrame {
     clauses: Vec<HeaderClause>,
@@ -60,82 +57,71 @@ impl FromIterator<HeaderClause> for HeaderFrame {
 
 impl IntoPy<HeaderFrame> for fastobo::ast::HeaderFrame {
     fn into_py(self, py: Python) -> HeaderFrame {
-        self
-            .into_iter()
-            .map(|clause| clause.into_py(py))
-            .collect()
+        self.into_iter().map(|clause| clause.into_py(py)).collect()
     }
 }
 
 impl IntoPy<obo::HeaderFrame> for HeaderFrame {
     fn into_py(self, py: Python) -> obo::HeaderFrame {
-        self
-            .clauses
+        self.clauses
             .into_iter()
             .map(|clause| clause.into_py(py))
             .collect()
     }
 }
 
-impl ToPyObject for HeaderFrame {
-    fn to_object(&self, py: Python) -> PyObject {
-        IntoPy::into_py(PyList::new(py, &self.clauses), py)
-    }
-}
+// impl ToPyObject for HeaderFrame {
+//     fn to_object(&self, py: Python) -> PyObject {
+//         IntoPy::into_py(PyList::new(py, &self.clauses), py)
+//     }
+// }
 
 #[listlike(field = "clauses", type = "HeaderClause")]
 #[pymethods]
 impl HeaderFrame {
     #[new]
-    pub fn __init__(clauses: Option<&PyAny>) -> PyResult<PyClassInitializer<Self>> {
+    #[pyo3(signature = (clauses = None))]
+    pub fn __init__<'py>(
+        clauses: Option<&Bound<'py, PyAny>>,
+    ) -> PyResult<PyClassInitializer<Self>> {
         let mut vec = Vec::new();
         if let Some(c) = clauses {
-            for item in PyIterator::from_object(c.py(), c)? {
-                vec.push(HeaderClause::extract(item?)?);
+            for item in PyIterator::from_object(c)? {
+                vec.push(item?.extract::<HeaderClause>()?);
             }
         }
         Ok(Self::new(vec).into())
     }
-}
 
-#[pyproto]
-impl PyObjectProtocol for HeaderFrame {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, HeaderFrame(self))
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, HeaderFrame(slf.clauses))
     }
 
     fn __str__(&self) -> PyResult<String> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let frame: obo::HeaderFrame = self.clone_py(py).into_py(py);
+        let frame: obo::HeaderFrame = Python::with_gil(|py| self.clone_py(py).into_py(py));
         Ok(frame.to_string())
     }
-}
 
-// FIXME(@althonos)
-#[pyproto]
-impl PySequenceProtocol for HeaderFrame {
     fn __len__(&self) -> PyResult<usize> {
         Ok(self.clauses.len())
     }
 
-    fn __getitem__(&self, index: isize) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
+    fn __getitem__(&self, index: isize) -> PyResult<Py<BaseHeaderClause>> {
         if index < self.clauses.len() as isize {
-            let item = &self.clauses[index as usize];
-            Ok(item.to_object(py))
+            Python::with_gil(|py| {
+                let item = &self.clauses[index as usize];
+                Ok(item.into_pyobject(py)?.unbind())
+            })
         } else {
             Err(PyIndexError::new_err("list index out of range"))
         }
     }
 
-    fn __setitem__(&mut self, index: isize, elem: &PyAny) -> PyResult<()> {
+    fn __setitem__<'py>(&mut self, index: isize, item: &Bound<'py, PyAny>) -> PyResult<()> {
         if index as usize > self.clauses.len() {
             return Err(PyIndexError::new_err("list index out of range"));
         }
-        let clause = HeaderClause::extract(elem)?;
+        let clause = item.extract::<HeaderClause>()?;
         self.clauses[index as usize] = clause;
         Ok(())
     }
@@ -148,17 +134,16 @@ impl PySequenceProtocol for HeaderFrame {
         Ok(())
     }
 
-    fn __concat__(&self, other: &PyAny) -> PyResult<Py<Self>> {
+    fn __concat__<'py>(&self, other: &Bound<'py, PyAny>) -> PyResult<Bound<'py, Self>> {
         let py = other.py();
 
-        let iterator = PyIterator::from_object(py, other)?;
+        let iterator = PyIterator::from_object(other)?;
         let mut new_clauses = self.clauses.clone_py(py);
         for item in iterator {
-            new_clauses.push(HeaderClause::extract(item?)?);
+            new_clauses.push(item?.extract::<HeaderClause>()?);
         }
 
-        let init = PyClassInitializer::from(AbstractFrame {})
-            .add_subclass(Self::new(new_clauses));
-        Py::new(py, init)
+        let init = PyClassInitializer::from(AbstractFrame {}).add_subclass(Self::new(new_clauses));
+        Bound::new(py, init)
     }
 }
