@@ -1,3 +1,4 @@
+use std::convert::Infallible;
 use std::fmt::Display;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
@@ -10,96 +11,98 @@ use pyo3::exceptions::PyValueError;
 use pyo3::prelude::*;
 use pyo3::types::PyAny;
 use pyo3::types::PyString;
-use pyo3::AsPyPointer;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
 use pyo3::PyTypeInfo;
 
 use fastobo::ast;
 use fastobo::parser::FromSlice;
 
-use crate::raise;
 use crate::error::Error;
-use crate::utils::ClonePy;
-use crate::utils::FinalClass;
+use crate::raise;
 use crate::utils::AbstractClass;
+use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::FinalClass;
+use crate::utils::IntoPy;
 
 // --- Module export ----------------------------------------------------------
 
+/// parse(s)
+/// --
+///
+/// Parse a string into an OBO identifier.
+///
+/// Arguments:
+///     s (`str`): the string representation of an OBO identifier
+///
+/// Returns:
+///     `~fastobo.id.BaseIdent`: the appropriate concrete subclass that
+///     can store the given identifier.
+///
+/// Raises:
+///     ValueError: when the string could not be parsed as a valid OBO
+///         identifier.
+///
+/// Example:
+///     >>> fastobo.id.parse("MS:1000031")
+///     PrefixedIdent('MS', '1000031')
+///     >>> fastobo.id.parse("part_of")
+///     UnprefixedIdent('part_of')
+///     >>> fastobo.id.parse("http://purl.obolibrary.org/obo/IAO_0000231")
+///     Url('http://purl.obolibrary.org/obo/IAO_0000231')
+///
+#[pyfunction]
+#[pyo3(name = "parse")]
+fn parse(py: Python, s: &str) -> PyResult<Ident> {
+    match fastobo::ast::Ident::from_str(s) {
+        Ok(id) => Ok(id.into_py(py)),
+        Err(e) => {
+            let err = PyErr::from(Error::from(e));
+            raise!(py, PyValueError("could not parse identifier") from err)
+        }
+    }
+}
+
+/// is_valid(s)
+/// --
+///
+/// Check whether or not a string is a valid OBO identifier.
+///
+/// Arguments
+///     s (`str`): the identifier to validate.
+///
+/// Returns:
+///     `bool`: whether or not the string is valid as an OBO identifier.
+///
+/// Example
+///     >>> fastobo.id.is_valid("MS:1000031")
+///     True
+///     >>> fastobo.id.is_valid("https://purl.obolibrary.org/obo/MS_1000031")
+///     True
+///     >>> fastobo.id.is_valid("related_to")
+///     True
+///     >>> fastobo.id.is_valid("definitely not an identifier")
+///     False
+#[pyfunction]
+#[pyo3(name = "is_valid")]
+fn is_valid(_py: Python, s: &str) -> bool {
+    let rule = fastobo::syntax::Rule::Id;
+    match fastobo::syntax::Lexer::tokenize(rule, s) {
+        Err(e) => false,
+        Ok(pairs) => pairs.as_str().len() == s.len(),
+    }
+}
+
 #[pymodule]
 #[pyo3(name = "id")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::BaseIdent>()?;
     m.add_class::<self::PrefixedIdent>()?;
     m.add_class::<self::UnprefixedIdent>()?;
     m.add_class::<self::Url>()?;
     m.add("__name__", "fastobo.id")?;
 
-    /// parse(s)
-    /// --
-    ///
-    /// Parse a string into an OBO identifier.
-    ///
-    /// Arguments:
-    ///     s (`str`): the string representation of an OBO identifier
-    ///
-    /// Returns:
-    ///     `~fastobo.id.BaseIdent`: the appropriate concrete subclass that
-    ///     can store the given identifier.
-    ///
-    /// Raises:
-    ///     ValueError: when the string could not be parsed as a valid OBO
-    ///         identifier.
-    ///
-    /// Example:
-    ///     >>> fastobo.id.parse("MS:1000031")
-    ///     PrefixedIdent('MS', '1000031')
-    ///     >>> fastobo.id.parse("part_of")
-    ///     UnprefixedIdent('part_of')
-    ///     >>> fastobo.id.parse("http://purl.obolibrary.org/obo/IAO_0000231")
-    ///     Url('http://purl.obolibrary.org/obo/IAO_0000231')
-    ///
-    #[pyfn(m)]
-    #[pyo3(name = "parse")]
-    fn parse(py: Python, s: &str) -> PyResult<Ident> {
-        match fastobo::ast::Ident::from_str(s) {
-            Ok(id) => Ok(id.into_py(py)),
-            Err(e) => {
-                let err = PyErr::from(Error::from(e));
-                raise!(py, PyValueError("could not parse identifier") from err)
-            }
-        }
-    }
-
-    /// is_valid(s)
-    /// --
-    ///
-    /// Check whether or not a string is a valid OBO identifier.
-    ///
-    /// Arguments
-    ///     s (`str`): the identifier to validate.
-    ///
-    /// Returns:
-    ///     `bool`: whether or not the string is valid as an OBO identifier.
-    ///
-    /// Example
-    ///     >>> fastobo.id.is_valid("MS:1000031")
-    ///     True
-    ///     >>> fastobo.id.is_valid("https://purl.obolibrary.org/obo/MS_1000031")
-    ///     True
-    ///     >>> fastobo.id.is_valid("related_to")
-    ///     True
-    ///     >>> fastobo.id.is_valid("definitely not an identifier")
-    ///     False
-    #[pyfn(m)]
-    #[pyo3(name = "is_valid")]
-    fn is_valid(_py: Python, s: &str) -> bool {
-        let rule = fastobo::syntax::Rule::Id;
-        match fastobo::syntax::Lexer::tokenize(rule, s) {
-            Err(e) => false,
-            Ok(pairs) => pairs.as_str().len() == s.len(),
-        }
-    }
+    m.add_function(wrap_pyfunction!(self::parse, m)?)?;
+    m.add_function(wrap_pyfunction!(self::is_valid, m)?)?;
 
     Ok(())
 }
@@ -124,7 +127,7 @@ macro_rules! impl_convert {
     };
 }
 
-#[derive(ClonePy, Debug, PartialEq, PyWrapper)]
+#[derive(Debug, ClonePy, PyWrapper, EqPy)]
 #[wraps(BaseIdent)]
 pub enum Ident {
     Unprefixed(Py<UnprefixedIdent>),
@@ -134,24 +137,34 @@ pub enum Ident {
 
 impl Display for Ident {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
-        match self {
-            Ident::Unprefixed(id) => id.as_ref(py).borrow().fmt(f),
-            Ident::Prefixed(id) => id.as_ref(py).borrow().fmt(f),
-            Ident::Url(id) => id.as_ref(py).borrow().fmt(f),
-        }
+        Python::with_gil(|py| match self {
+            Ident::Unprefixed(id) => id.bind(py).borrow().fmt(f),
+            Ident::Prefixed(id) => id.bind(py).borrow().fmt(f),
+            Ident::Url(id) => id.bind(py).borrow().fmt(f),
+        })
     }
 }
 
+// impl EqPy for Ident {
+//     fn eq_py(&self, other: &Self, py: Python) -> bool {
+//         use self::Ident::*;
+//         match (&self, &other) {
+//             (Unprefixed(x), Unprefixed(y)) => x.bind(py).eq(y.bind(py)).unwrap(),
+//             (Prefixed(x), Prefixed(y)) => x.bind(py).eq(y.bind(py)).unwrap(),
+//             (Url(x), Url(y)) => x.bind(py).eq(y.bind(py)).unwrap(),
+//             _ => false,
+//         }
+//     }
+// }
+
 impl IntoPy<Ident> for fastobo::ast::Ident {
     fn into_py(self, py: Python) -> Ident {
         match self {
             ast::Ident::Unprefixed(id) => Py::new(py, id.into_py(py)).map(Ident::Unprefixed),
             ast::Ident::Prefixed(id) => Py::new(py, id.into_py(py)).map(Ident::Prefixed),
             ast::Ident::Url(id) => Py::new(py, id.into_py(py)).map(Ident::Url),
-        }.expect("could not allocate on Python heap")
+        }
+        .expect("could not allocate on Python heap")
     }
 }
 
@@ -159,15 +172,15 @@ impl IntoPy<fastobo::ast::Ident> for Ident {
     fn into_py(self, py: Python) -> fastobo::ast::Ident {
         match self {
             Ident::Unprefixed(id) => {
-                let i = id.as_ref(py).borrow();
+                let i = id.bind(py).borrow();
                 ast::Ident::from((*i).inner.clone())
             }
             Ident::Prefixed(id) => {
-                let i = id.as_ref(py).borrow();
+                let i = id.bind(py).borrow();
                 ast::Ident::from((*i).inner.clone())
             }
             Ident::Url(id) => {
-                let i = id.as_ref(py).borrow();
+                let i = id.bind(py).borrow();
                 ast::Ident::from((*i).inner.clone())
             }
         }
@@ -189,7 +202,7 @@ impl_convert!(NamespaceIdent, Ident);
 pub struct BaseIdent {}
 
 impl AbstractClass for BaseIdent {
-    fn initializer() -> PyClassInitializer<Self> {
+    fn initializer(py: Python) -> PyClassInitializer<Self> {
         PyClassInitializer::from(BaseIdent {})
     }
 }
@@ -208,7 +221,7 @@ impl AbstractClass for BaseIdent {
 ///     'GO:0009637'
 ///
 #[pyclass(extends=BaseIdent, module="fastobo.id")]
-#[derive(Debug, FinalClass, Clone, PartialEq, Eq)]
+#[derive(Debug, FinalClass, Clone, PartialEq, Eq, EqPy)]
 #[base(BaseIdent)]
 pub struct PrefixedIdent {
     inner: ast::PrefixedIdent,
@@ -217,7 +230,7 @@ pub struct PrefixedIdent {
 impl PrefixedIdent {
     fn new(prefix: &str, local: &str) -> Self {
         PrefixedIdent {
-            inner: ast::PrefixedIdent::new(prefix, local)
+            inner: ast::PrefixedIdent::new(prefix, local),
         }
     }
 }
@@ -236,9 +249,7 @@ impl Display for PrefixedIdent {
 
 impl From<ast::PrefixedIdent> for PrefixedIdent {
     fn from(id: ast::PrefixedIdent) -> Self {
-        Self {
-            inner: id
-        }
+        Self { inner: id }
     }
 }
 
@@ -262,9 +273,7 @@ impl IntoPy<ast::Ident> for PrefixedIdent {
 
 impl IntoPy<PrefixedIdent> for ast::PrefixedIdent {
     fn into_py(self, _py: Python) -> PrefixedIdent {
-        PrefixedIdent {
-            inner: self
-        }
+        PrefixedIdent { inner: self }
     }
 }
 
@@ -281,59 +290,25 @@ impl PrefixedIdent {
     ///
     #[new]
     fn __init__(prefix: &str, local: &str) -> PyResult<PyClassInitializer<Self>> {
-        Ok(
-            PyClassInitializer::from(BaseIdent {})
-                .add_subclass(Self::new(prefix, local))
-        )
+        Ok(PyClassInitializer::from(BaseIdent {}).add_subclass(Self::new(prefix, local)))
     }
 
-    /// `~fastobo.id.IdentPrefix`: the IDspace of the identifier.
-    #[getter]
-    fn get_prefix<'py>(&self) -> &str {
-        self.inner.prefix()
-    }
-
-    #[setter]
-    fn set_prefix(&mut self, prefix: &str) {
-        self.inner = ast::PrefixedIdent::new(
-            prefix,
-            self.inner.local()
-        );
-    }
-
-    /// `~fastobo.id.IdentLocal`: the local part of the identifier.
-    #[getter]
-    fn get_local<'py>(&self) -> &str {
-        self.inner.local()
-    }
-
-    #[setter]
-    fn set_local(&mut self, local: &str) {
-        self.inner = ast::PrefixedIdent::new(
-            self.inner.prefix(),
-            local
-        );
-    }
-}
-
-#[pyproto]
-impl PyObjectProtocol for PrefixedIdent {
     fn __hash__(&self) -> u64 {
         impl_hash!(self.inner.prefix(), ":", self.inner.local())
     }
 
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, PrefixedIdent(self.inner.prefix(), self.inner.local()))
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, PrefixedIdent(slf.inner.prefix(), slf.inner.local()))
     }
 
     fn __str__(&self) -> PyResult<String> {
         Ok(self.to_string())
     }
 
-    fn __richcmp__(&self, other: &PyAny, op: CompareOp) -> PyResult<bool> {
+    fn __richcmp__<'py>(&self, other: &Bound<'py, PyAny>, op: CompareOp) -> PyResult<bool> {
         let py = other.py();
-        if let Ok(r) = other.extract::<Py<PrefixedIdent>>() {
-            let r = r.as_ref(py).borrow();
+        if let Ok(r) = other.downcast::<PrefixedIdent>() {
+            let r = r.as_borrowed().borrow();
             let lp = self.inner.prefix();
             let ll = self.inner.local();
             let rp = r.inner.prefix();
@@ -358,6 +333,28 @@ impl PyObjectProtocol for PrefixedIdent {
             }
         }
     }
+
+    /// `str`: the IDspace of the identifier.
+    #[getter]
+    fn get_prefix<'py>(&self) -> &str {
+        self.inner.prefix()
+    }
+
+    #[setter]
+    fn set_prefix(&mut self, prefix: &str) {
+        self.inner = ast::PrefixedIdent::new(prefix, self.inner.local());
+    }
+
+    /// `str`: the local part of the identifier.
+    #[getter]
+    fn get_local<'py>(&self) -> &str {
+        self.inner.local()
+    }
+
+    #[setter]
+    fn set_local(&mut self, local: &str) {
+        self.inner = ast::PrefixedIdent::new(self.inner.prefix(), local);
+    }
 }
 
 // --- UnprefixedIdent --------------------------------------------------------
@@ -373,7 +370,7 @@ impl PyObjectProtocol for PrefixedIdent {
 ///     hello world
 ///
 #[pyclass(extends=BaseIdent, module="fastobo.id")]
-#[derive(Clone, Debug, Eq, Hash, PartialEq, FinalClass)]
+#[derive(Clone, Debug, Eq, Hash, PartialEq, EqPy, FinalClass)]
 #[base(BaseIdent)]
 pub struct UnprefixedIdent {
     inner: ast::UnprefixedIdent,
@@ -448,47 +445,31 @@ impl UnprefixedIdent {
     #[new]
     fn __init__(value: &str) -> PyClassInitializer<Self> {
         let id = ast::UnprefixedIdent::new(value.to_string());
-        PyClassInitializer::from(BaseIdent {})
-            .add_subclass(UnprefixedIdent::new(id))
+        PyClassInitializer::from(BaseIdent {}).add_subclass(UnprefixedIdent::new(id))
     }
 
-    /// `str`: the escaped representation of the identifier.
-    #[getter]
-    fn escaped(&self) -> PyResult<String> {
-        Ok(self.inner.to_string())
-    }
-
-    /// `str`: the unescaped representation of the identifier.
-    #[getter]
-    fn unescaped(&self) -> PyResult<&str> {
-        Ok(self.inner.as_str())
-    }
-}
-
-#[pyproto]
-impl PyObjectProtocol for UnprefixedIdent {
     fn __hash__(&self) -> u64 {
         impl_hash!(self.inner.as_str())
     }
 
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, UnprefixedIdent(self.inner.as_str()))
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, UnprefixedIdent(slf.inner.as_str()))
     }
 
-    fn __str__(&'p self) -> PyResult<&'p str> {
+    fn __str__(&self) -> PyResult<&str> {
         Ok(self.inner.as_str())
     }
 
-    fn __richcmp__(&self, other: &PyAny, op: CompareOp) -> PyResult<bool> {
-        if let Ok(u) = other.extract::<Py<UnprefixedIdent>>() {
-            let u = u.as_ref(other.py()).borrow();
+    fn __richcmp__<'py>(&self, other: &Bound<'py, PyAny>, op: CompareOp) -> PyResult<bool> {
+        if let Ok(u) = other.downcast::<UnprefixedIdent>() {
+            let x = u.as_borrowed().borrow();
             match op {
-                CompareOp::Lt => Ok(self.inner < u.inner),
-                CompareOp::Le => Ok(self.inner <= u.inner),
-                CompareOp::Eq => Ok(self.inner == u.inner),
-                CompareOp::Ne => Ok(self.inner != u.inner),
-                CompareOp::Gt => Ok(self.inner > u.inner),
-                CompareOp::Ge => Ok(self.inner >= u.inner),
+                CompareOp::Lt => Ok(self.inner < x.inner),
+                CompareOp::Le => Ok(self.inner <= x.inner),
+                CompareOp::Eq => Ok(self.inner == x.inner),
+                CompareOp::Ne => Ok(self.inner != x.inner),
+                CompareOp::Gt => Ok(self.inner > x.inner),
+                CompareOp::Ge => Ok(self.inner >= x.inner),
             }
         } else {
             match op {
@@ -502,6 +483,18 @@ impl PyObjectProtocol for UnprefixedIdent {
             }
         }
     }
+
+    /// `str`: the escaped representation of the identifier.
+    #[getter]
+    fn escaped(&self) -> PyResult<String> {
+        Ok(self.inner.to_string())
+    }
+
+    /// `str`: the unescaped representation of the identifier.
+    #[getter]
+    fn unescaped(&self) -> PyResult<&str> {
+        Ok(self.inner.as_str())
+    }
 }
 
 // --- UrlIdent ---------------------------------------------------------------
@@ -521,7 +514,7 @@ impl PyObjectProtocol for UnprefixedIdent {
 ///     ValueError: invalid url: ...
 ///
 #[pyclass(extends=BaseIdent, module="fastobo.id")]
-#[derive(Clone, ClonePy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd, FinalClass)]
+#[derive(Clone, ClonePy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd, EqPy, FinalClass)]
 #[base(BaseIdent)]
 pub struct Url {
     inner: ast::Url,
@@ -545,7 +538,6 @@ impl From<ast::Url> for Url {
     }
 }
 
-
 impl From<Url> for fastobo::ast::Url {
     fn from(url: Url) -> Self {
         url.inner
@@ -594,34 +586,31 @@ impl Url {
             Err(e) => Err(PyValueError::new_err(format!("invalid url: {}", e))),
         }
     }
-}
 
-#[pyproto]
-impl PyObjectProtocol for Url {
     fn __hash__(&self) -> u64 {
         impl_hash!(self.inner.as_str())
     }
 
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, Url(self.inner.as_str()))
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, Url(slf.inner.as_str()))
     }
 
     /// Retrieve the URL in a serialized form.
-    fn __str__(&'p self) -> PyResult<&'p str> {
+    fn __str__(&self) -> PyResult<&str> {
         Ok(self.inner.as_str())
     }
 
-    /// Compare to another `Url` or `str` instance.
-    fn __richcmp__(&self, other: &PyAny, op: CompareOp) -> PyResult<bool> {
-        if let Ok(url) = other.extract::<Py<Url>>() {
-            let url = &*url.as_ref(other.py()).borrow();
+    /// Compare to another `Url` instance.
+    fn __richcmp__<'py>(&self, other: Bound<'py, PyAny>, op: CompareOp) -> PyResult<bool> {
+        if let Ok(url) = other.downcast::<Url>() {
+            let x = url.as_borrowed().borrow();
             match op {
-                CompareOp::Lt => Ok(self < url),
-                CompareOp::Le => Ok(self <= url),
-                CompareOp::Eq => Ok(self == url),
-                CompareOp::Ne => Ok(self != url),
-                CompareOp::Gt => Ok(self > url),
-                CompareOp::Ge => Ok(self >= url),
+                CompareOp::Lt => Ok(self < &*x),
+                CompareOp::Le => Ok(self <= &*x),
+                CompareOp::Eq => Ok(self == &*x),
+                CompareOp::Ne => Ok(self != &*x),
+                CompareOp::Gt => Ok(self > &*x),
+                CompareOp::Ge => Ok(self >= &*x),
             }
         } else {
             match op {
