@@ -8,22 +8,20 @@ use pyo3::prelude::*;
 use pyo3::types::PyAny;
 use pyo3::types::PyIterator;
 use pyo3::types::PyString;
-use pyo3::AsPyPointer;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use fastobo::ast;
 
 use super::super::abc::AbstractEntityFrame;
 use super::super::id::Ident;
-use crate::utils::ClonePy;
-use crate::utils::FinalClass;
 use crate::utils::AbstractClass;
+use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::FinalClass;
+use crate::utils::IntoPy;
 
 #[pyclass(extends=AbstractEntityFrame, module="fastobo.instance")]
-#[derive(Debug, FinalClass)]
+#[derive(Debug, FinalClass, EqPy)]
 #[base(AbstractEntityFrame)]
 pub struct InstanceFrame {
     id: Ident,
@@ -52,9 +50,8 @@ impl ClonePy for InstanceFrame {
 
 impl Display for InstanceFrame {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let frame: fastobo::ast::InstanceFrame = self.clone_py(py).into_py(py);
+        let frame: fastobo::ast::InstanceFrame =
+            Python::with_gil(|py| self.clone_py(py).into_py(py));
         frame.fmt(f)
     }
 }
@@ -68,9 +65,7 @@ impl IntoPy<InstanceFrame> for fastobo::ast::InstanceFrame {
 
 impl IntoPy<fastobo::ast::InstanceFrame> for InstanceFrame {
     fn into_py(self, py: Python) -> fastobo::ast::InstanceFrame {
-        fastobo::ast::InstanceFrame::new(
-            fastobo::ast::InstanceIdent::new(self.id.into_py(py))
-        )
+        fastobo::ast::InstanceFrame::new(fastobo::ast::InstanceIdent::new(self.id.into_py(py)))
     }
 }
 
