@@ -3,8 +3,9 @@ pub mod frame;
 
 use pyo3::prelude::*;
 
-#[pymodule(instance)]
-pub fn init(py: Python, m: &PyModule) -> PyResult<()> {
+#[pymodule]
+#[pyo3(name = "instance")]
+pub fn init<'py>(py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::frame::InstanceFrame>()?;
 
     register!(py, m, InstanceFrame, "collections.abc", MutableSequence);
