@@ -8,22 +8,22 @@ use pyo3::exceptions::PyTypeError;
 use pyo3::prelude::*;
 use pyo3::types::PyAny;
 use pyo3::types::PyString;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
 use pyo3::PyTypeInfo;
 
 use fastobo::ast;
 
 use super::id::Ident;
-use crate::utils::ClonePy;
-use crate::utils::FinalClass;
 use crate::utils::AbstractClass;
+use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::FinalClass;
+use crate::utils::IntoPy;
 
 // --- Module export ---------------------------------------------------------
 
 #[pymodule]
 #[pyo3(name = "pv")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::AbstractPropertyValue>()?;
     m.add_class::<self::LiteralPropertyValue>()?;
     m.add_class::<self::ResourcePropertyValue>()?;
@@ -33,7 +33,7 @@ pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
 
 // --- Conversion Wrapper ----------------------------------------------------
 
-#[derive(ClonePy, Debug, PartialEq, PyWrapper)]
+#[derive(ClonePy, Debug, PyWrapper, EqPy)]
 #[wraps(AbstractPropertyValue)]
 pub enum PropertyValue {
     Literal(Py<LiteralPropertyValue>),
@@ -42,12 +42,10 @@ pub enum PropertyValue {
 
 impl Display for PropertyValue {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        match self {
-            PropertyValue::Literal(lpv) => lpv.as_ref(py).borrow().fmt(f),
-            PropertyValue::Resource(rpv) => rpv.as_ref(py).borrow().fmt(f),
-        }
+        Python::with_gil(|py| match self {
+            PropertyValue::Literal(lpv) => lpv.bind(py).borrow().fmt(f),
+            PropertyValue::Resource(rpv) => rpv.bind(py).borrow().fmt(f),
+        })
     }
 }
 
@@ -68,12 +66,8 @@ impl IntoPy<PropertyValue> for fastobo::ast::PropertyValue {
 impl IntoPy<fastobo::ast::PropertyValue> for PropertyValue {
     fn into_py(self, py: Python) -> fastobo::ast::PropertyValue {
         match self {
-            PropertyValue::Literal(t) => {
-                t.as_ref(py).borrow().deref().clone_py(py).into_py(py)
-            }
-            PropertyValue::Resource(r) => {
-                r.as_ref(py).borrow().deref().clone_py(py).into_py(py)
-            }
+            PropertyValue::Literal(t) => t.bind(py).borrow().deref().clone_py(py).into_py(py),
+            PropertyValue::Resource(r) => r.bind(py).borrow().deref().clone_py(py).into_py(py),
         }
     }
 }
@@ -85,7 +79,7 @@ impl IntoPy<fastobo::ast::PropertyValue> for PropertyValue {
 pub struct AbstractPropertyValue {}
 
 impl AbstractClass for AbstractPropertyValue {
-    fn initializer() -> PyClassInitializer<Self> {
+    fn initializer(py: Python) -> PyClassInitializer<Self> {
         PyClassInitializer::from(Self {})
     }
 }
@@ -93,7 +87,7 @@ impl AbstractClass for AbstractPropertyValue {
 // --- Literal -----------------------------------------------------------------
 
 #[pyclass(extends=AbstractPropertyValue, module="fastobo.pv")]
-#[derive(Debug, FinalClass)]
+#[derive(Debug, FinalClass, EqPy)]
 #[base(AbstractPropertyValue)]
 pub struct LiteralPropertyValue {
     relation: Ident,
@@ -103,7 +97,11 @@ pub struct LiteralPropertyValue {
 
 impl LiteralPropertyValue {
     pub fn new(relation: Ident, value: fastobo::ast::QuotedString, datatype: Ident) -> Self {
-        LiteralPropertyValue { relation, value, datatype }
+        LiteralPropertyValue {
+            relation,
+            value,
+            datatype,
+        }
     }
 }
 
@@ -119,9 +117,7 @@ impl ClonePy for LiteralPropertyValue {
 
 impl Display for LiteralPropertyValue {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let pv: fastobo::ast::PropertyValue = self.clone_py(py).into_py(py);
+        let pv: fastobo::ast::PropertyValue = Python::with_gil(|py| self.clone_py(py).into_py(py));
         pv.fmt(f)
     }
 }
@@ -131,7 +127,7 @@ impl IntoPy<fastobo::ast::LiteralPropertyValue> for LiteralPropertyValue {
         fastobo::ast::LiteralPropertyValue::new(
             self.relation.into_py(py),
             self.value,
-            self.datatype.into_py(py)
+            self.datatype.into_py(py),
         )
     }
 }
@@ -154,13 +150,13 @@ impl IntoPy<LiteralPropertyValue> for fastobo::ast::LiteralPropertyValue {
 #[pymethods]
 impl LiteralPropertyValue {
     #[new]
-    fn __init__(
-        relation: &PyAny,
-        value: &PyAny,
-        datatype: &PyAny,
+    fn __init__<'py>(
+        relation: Bound<'py, PyAny>,
+        value: Bound<'py, PyAny>,
+        datatype: Bound<'py, PyAny>,
     ) -> PyResult<PyClassInitializer<Self>> {
         let r = relation.extract::<Ident>()?;
-        let v = if let Ok(s) = value.extract::<&PyString>() {
+        let v = if let Ok(s) = value.downcast::<PyString>() {
             ast::QuotedString::new(s.to_str()?.to_string())
         } else {
             let n = value.get_type().name()?;
@@ -171,6 +167,19 @@ impl LiteralPropertyValue {
         Ok(Self::new(r, v, dt).into())
     }
 
+    fn __repr__<'py>(&self, py: Python<'py>) -> PyResult<Bound<'py, PyAny>> {
+        let fmt = PyString::new(py, "LiteralPropertyValue({!r}, {!r}, {!r})");
+        fmt.call_method1(
+            "format",
+            (&self.relation, self.value.as_str(), &self.datatype),
+        )
+    }
+
+    fn __str__(&self) -> PyResult<String> {
+        let pv: fastobo::ast::PropertyValue = Python::with_gil(|py| self.clone_py(py).into_py(py));
+        Ok(pv.to_string())
+    }
+
     #[getter]
     fn get_relation(&self) -> PyResult<&Ident> {
         Ok(&self.relation)
@@ -205,35 +214,10 @@ impl LiteralPropertyValue {
     }
 }
 
-#[pyproto]
-impl PyObjectProtocol for LiteralPropertyValue {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let fmt = PyString::new(py, "LiteralPropertyValue({!r}, {!r}, {!r})");
-        fmt.to_object(py).call_method1(
-            py,
-            "format",
-            (
-                self.relation.to_object(py),
-                self.value.as_str(),
-                self.datatype.to_object(py),
-            ),
-        )
-    }
-
-    fn __str__(&self) -> PyResult<String> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let pv: fastobo::ast::PropertyValue = self.clone_py(py).into_py(py);
-        Ok(pv.to_string())
-    }
-}
-
 // --- Resource ------------------------------------------------------------
 
 #[pyclass(extends=AbstractPropertyValue, module="fastobo.pv")]
-#[derive(Debug, FinalClass)]
+#[derive(Debug, FinalClass, EqPy)]
 #[base(AbstractPropertyValue)]
 pub struct ResourcePropertyValue {
     relation: Ident,
@@ -242,10 +226,7 @@ pub struct ResourcePropertyValue {
 
 impl ResourcePropertyValue {
     pub fn new(relation: Ident, value: Ident) -> Self {
-        ResourcePropertyValue {
-            relation,
-            value,
-        }
+        ResourcePropertyValue { relation, value }
     }
 }
 
@@ -260,10 +241,7 @@ impl ClonePy for ResourcePropertyValue {
 
 impl Display for ResourcePropertyValue {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
-        let pv: fastobo::ast::PropertyValue = self.clone_py(py).into_py(py);
+        let pv: fastobo::ast::PropertyValue = Python::with_gil(|py| self.clone_py(py).into_py(py));
         pv.fmt(f)
     }
 }
@@ -292,10 +270,18 @@ impl IntoPy<ResourcePropertyValue> for fastobo::ast::ResourcePropertyValue {
 impl ResourcePropertyValue {
     #[new]
     fn __init__(relation: Ident, value: Ident) -> PyClassInitializer<Self> {
-        let gil = Python::acquire_gil();
         Self::new(relation, value).into()
     }
 
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, ResourcePropertyValue(slf.relation, slf.value))
+    }
+
+    fn __str__(&self) -> PyResult<String> {
+        let pv: fastobo::ast::PropertyValue = Python::with_gil(|py| self.clone_py(py).into_py(py));
+        Ok(pv.to_string())
+    }
+
     #[getter]
     fn get_relation(&self) -> PyResult<&Ident> {
         Ok(&self.relation)
@@ -318,17 +304,3 @@ impl ResourcePropertyValue {
         Ok(())
     }
 }
-
-#[pyproto]
-impl PyObjectProtocol for ResourcePropertyValue {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, ResourcePropertyValue(self.relation, self.value))
-    }
-
-    fn __str__(&self) -> PyResult<String> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let pv: fastobo::ast::PropertyValue = self.clone_py(py).into_py(py);
-        Ok(pv.to_string())
-    }
-}
