@@ -1,3 +1,4 @@
+use std::convert::Infallible;
 use std::fmt::Display;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
@@ -14,32 +15,28 @@ use pyo3::types::PyAny;
 use pyo3::types::PyIterator;
 use pyo3::types::PyList;
 use pyo3::types::PyString;
-use pyo3::AsPyPointer;
-use pyo3::PyGCProtocol;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use super::id::Ident;
 use super::xref::XrefList;
 use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::IntoPy;
 
 // --- Module export ---------------------------------------------------------
 
 #[pymodule]
 #[pyo3(name = "syn")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::Synonym>()?;
-    m.add_class::<self::SynonymScope>()?;
     m.add("__name__", "fastobo.syn")?;
     Ok(())
 }
 
 // --- SynonymScope ----------------------------------------------------------
 
-#[pyclass(module = "fastobo.syn")] // FIXME(@althonos): probably not needed since it is not exposed.
-#[derive(Clone, ClonePy, Debug, Eq, PartialEq)]
+// #[pyclass(module = "fastobo.syn")] // FIXME(@althonos): probably not needed since it is not exposed.
+#[derive(Clone, ClonePy, Debug, Eq, PartialEq, EqPy)]
 pub struct SynonymScope {
     inner: fastobo::ast::SynonymScope,
 }
@@ -76,14 +73,10 @@ impl FromStr for SynonymScope {
             "BROAD" => Ok(Self::new(fastobo::ast::SynonymScope::Broad)),
             "NARROW" => Ok(Self::new(fastobo::ast::SynonymScope::Narrow)),
             "RELATED" => Ok(Self::new(fastobo::ast::SynonymScope::Related)),
-            invalid => {
-                Err(PyValueError::new_err(
-                    format!(
-                        "expected 'EXACT', 'BROAD', 'NARROW' or 'RELATED', found {:?}",
-                        invalid
-                    )
-                ))
-            }
+            invalid => Err(PyValueError::new_err(format!(
+                "expected 'EXACT', 'BROAD', 'NARROW' or 'RELATED', found {:?}",
+                invalid
+            ))),
         }
     }
 }
@@ -100,16 +93,33 @@ impl IntoPy<fastobo::ast::SynonymScope> for SynonymScope {
     }
 }
 
-impl ToPyObject for SynonymScope {
-    fn to_object(&self, py: Python) -> PyObject {
-        self.to_string().to_object(py)
+impl<'py> IntoPyObject<'py> for &SynonymScope {
+    type Error = Infallible;
+    type Target = PyString;
+    type Output = Bound<'py, PyString>;
+    fn into_pyobject(self, py: Python<'py>) -> Result<Self::Output, Self::Error> {
+        match self.inner {
+            fastobo::ast::SynonymScope::Exact => Ok(pyo3::intern!(py, "EXACT").clone()),
+            fastobo::ast::SynonymScope::Broad => Ok(pyo3::intern!(py, "BROAD").clone()),
+            fastobo::ast::SynonymScope::Narrow => Ok(pyo3::intern!(py, "NARROW").clone()),
+            fastobo::ast::SynonymScope::Related => Ok(pyo3::intern!(py, "RELATED").clone()),
+        }
+    }
+}
+
+impl<'py> IntoPyObject<'py> for SynonymScope {
+    type Error = Infallible;
+    type Target = PyString;
+    type Output = Bound<'py, PyString>;
+    fn into_pyobject(self, py: Python<'py>) -> Result<Self::Output, Self::Error> {
+        (&self).into_pyobject(py)
     }
 }
 
 // --- Synonym ---------------------------------------------------------------
 
 #[pyclass(module = "fastobo.syn")]
-#[derive(Debug, PartialEq)]
+#[derive(Debug, EqPy)]
 pub struct Synonym {
     desc: fastobo::ast::QuotedString,
     scope: SynonymScope,
@@ -131,9 +141,7 @@ impl ClonePy for Synonym {
 
 impl Display for Synonym {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let syn: fastobo::ast::Synonym = self.clone_py(py).into_py(py);
+        let syn: fastobo::ast::Synonym = Python::with_gil(|py| self.clone_py(py).into_py(py));
         syn.fmt(f)
     }
 }
@@ -144,10 +152,8 @@ impl IntoPy<Synonym> for fastobo::ast::Synonym {
             desc: std::mem::take(self.description_mut()),
             scope: SynonymScope::new(self.scope().clone()),
             ty: self.ty().map(|id| id.clone().into_py(py)),
-            xrefs: Py::new(
-                py,
-                std::mem::take(self.xrefs_mut()).into_py(py),
-            ).expect("failed allocating memory on Python heap")
+            xrefs: Py::new(py, std::mem::take(self.xrefs_mut()).into_py(py))
+                .expect("failed allocating memory on Python heap"),
         }
     }
 }
@@ -158,37 +164,53 @@ impl IntoPy<fastobo::ast::Synonym> for Synonym {
             self.desc,
             self.scope.inner,
             self.ty.map(|ty| ty.into_py(py)),
-            (&*self.xrefs.as_ref(py).borrow()).into_py(py)
+            (&*self.xrefs.bind(py).borrow()).into_py(py),
         )
     }
 }
 
 #[pymethods]
 impl Synonym {
-
     #[new]
-    pub fn __init__(
+    #[pyo3(signature = (desc, scope, ty = None, xrefs = None))]
+    pub fn __init__<'py>(
         desc: String,
         scope: &str,
         ty: Option<Ident>,
-        xrefs: Option<&PyAny>
+        xrefs: Option<Bound<'py, PyAny>>,
     ) -> PyResult<Self> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-
-        let list = xrefs
-            .map(|x| XrefList::collect(py, x))
-            .transpose()?
-            .unwrap_or_default();
-
+        let xrefs = Python::with_gil(|py| {
+            let list = xrefs
+                .map(|x| XrefList::collect(py, &x))
+                .transpose()?
+                .unwrap_or_default();
+            Py::new(py, list)
+        })?;
         Ok(Self {
             desc: fastobo::ast::QuotedString::new(desc),
             scope: SynonymScope::from_str(scope)?,
-            xrefs: Py::new(py, list)?,
+            xrefs,
             ty,
         })
     }
 
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        impl_repr_py!(slf, Synonym(slf.desc, slf.scope))
+    }
+
+    fn __str__(&self) -> PyResult<String> {
+        Ok(self.to_string())
+    }
+
+    fn __richcmp__<'py>(&self, other: &Bound<'py, PyAny>, op: CompareOp) -> PyResult<PyObject> {
+        impl_richcmp_py!(
+            self,
+            other,
+            op,
+            self.desc && self.scope && self.ty && self.xrefs
+        )
+    }
+
     #[getter]
     pub fn get_desc(&self) -> PyResult<String> {
         Ok(self.desc.as_str().to_owned())
@@ -222,23 +244,3 @@ impl Synonym {
         Ok(())
     }
 }
-
-#[pyproto]
-impl PyObjectProtocol for Synonym {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        impl_repr!(self, Synonym(self.desc, self.scope))
-    }
-
-    fn __str__(&self) -> PyResult<String> {
-        Ok(self.to_string())
-    }
-
-    fn __richcmp__(&self, other: &PyAny, op: CompareOp) -> PyResult<PyObject> {
-        impl_richmp!(
-            self,
-            other,
-            op,
-            self.desc && self.scope && self.ty && self.xrefs
-        )
-    }
-}
