@@ -1,3 +1,4 @@
+use std::convert::Infallible;
 use std::fmt::Display;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
@@ -5,6 +6,7 @@ use std::rc::Rc;
 use std::str::FromStr;
 use std::string::ToString;
 
+use pyo3::class::basic::CompareOp;
 use pyo3::class::gc::PyVisit;
 use pyo3::exceptions::PyIndexError;
 use pyo3::exceptions::PyTypeError;
@@ -14,21 +16,18 @@ use pyo3::types::PyAny;
 use pyo3::types::PyIterator;
 use pyo3::types::PyList;
 use pyo3::types::PyString;
-use pyo3::AsPyPointer;
-use pyo3::PyGCProtocol;
-use pyo3::PyNativeType;
-use pyo3::PyObjectProtocol;
-use pyo3::PySequenceProtocol;
 use pyo3::PyTypeInfo;
 
 use super::id::Ident;
 use crate::utils::ClonePy;
+use crate::utils::EqPy;
+use crate::utils::IntoPy;
 
 // --- Module export ---------------------------------------------------------
 
 #[pymodule]
 #[pyo3(name = "xref")]
-pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
+pub fn init<'py>(_py: Python<'py>, m: &Bound<'py, PyModule>) -> PyResult<()> {
     m.add_class::<self::Xref>()?;
     m.add_class::<self::XrefList>()?;
     m.add("__name__", "fastobo.xref")?;
@@ -48,7 +47,7 @@ pub fn init(_py: Python, m: &PyModule) -> PyResult<()> {
 ///     ...     fastobo.id.PrefixedIdent('ISBN', '978-0-321-84268-8'),
 ///     ... )
 #[pyclass(module = "fastobo.xref")]
-#[derive(Debug, PartialEq)]
+#[derive(Debug, EqPy)]
 pub struct Xref {
     #[pyo3(set)]
     id: Ident,
@@ -76,9 +75,7 @@ impl ClonePy for Xref {
 
 impl Display for Xref {
     fn fmt(&self, f: &mut Formatter) -> FmtResult {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let xref: fastobo::ast::Xref = self.clone_py(py).into_py(py);
+        let xref: fastobo::ast::Xref = Python::with_gil(|py| self.clone_py(py).into_py(py));
         xref.fmt(f)
     }
 }
@@ -111,15 +108,31 @@ impl Xref {
     ///     id (~fastobo.id.Ident): the identifier of the reference.
     ///     desc (str, optional): an optional description for the reference.
     #[new]
+    #[pyo3(signature = (id, desc = None))]
     fn __init__(id: Ident, desc: Option<String>) -> Self {
         if let Some(s) = desc {
-            let gil = Python::acquire_gil();
             Self::with_desc(id, Some(fastobo::ast::QuotedString::new(s)))
         } else {
             Self::new(id)
         }
     }
 
+    fn __repr__<'py>(&self, py: Python<'py>) -> PyResult<Bound<'py, PyAny>> {
+        if let Some(ref d) = self.desc {
+            PyString::intern(py, "Xref({!r}, {!r})").call_method1("format", (&self.id, d.as_str()))
+        } else {
+            PyString::intern(py, "Xref({!r})").call_method1("format", (&self.id,))
+        }
+    }
+
+    fn __str__(&self) -> PyResult<String> {
+        Ok(self.to_string())
+    }
+
+    fn __richcmp__<'py>(&self, other: &Bound<'py, PyAny>, op: CompareOp) -> PyResult<PyObject> {
+        impl_richcmp_py!(self, other, op, self.id && self.desc)
+    }
+
     /// `~fastobo.id.Ident`: the identifier of the reference.
     #[getter]
     fn get_id(&self) -> PyResult<&Ident> {
@@ -142,27 +155,6 @@ impl Xref {
     }
 }
 
-#[pyproto]
-impl PyObjectProtocol for Xref {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        if let Some(ref d) = self.desc {
-            PyString::new(py, "Xref({!r}, {!r})")
-                .to_object(py)
-                .call_method1(py, "format", (&self.id, d.as_str()))
-        } else {
-            PyString::new(py, "Xref({!r})")
-                .to_object(py)
-                .call_method1(py, "format", (&self.id,))
-        }
-    }
-
-    fn __str__(&self) -> PyResult<String> {
-        Ok(self.to_string())
-    }
-}
-
 // --- XrefList --------------------------------------------------------------
 
 /// A list of cross-references.
@@ -175,7 +167,7 @@ impl PyObjectProtocol for Xref {
 ///     Xref(PrefixedIdent('PSI', 'MS'))
 ///
 #[pyclass(module = "fastobo.xref")]
-#[derive(Debug, Default, PartialEq)]
+#[derive(Debug, Default, EqPy)]
 pub struct XrefList {
     xrefs: Vec<Py<Xref>>,
 }
@@ -187,9 +179,9 @@ impl XrefList {
     }
 
     /// Create a new `XrefList` from a `PyIterator`.
-    pub fn collect(py: Python, xrefs: &PyAny) -> PyResult<Self> {
+    pub fn collect<'py>(py: Python<'py>, xrefs: &Bound<'py, PyAny>) -> PyResult<Self> {
         let mut vec = Vec::new();
-        for item in PyIterator::from_object(py, xrefs)? {
+        for item in PyIterator::from_object(xrefs)? {
             let i = item?;
             if let Ok(xref) = i.extract::<Py<Xref>>() {
                 vec.push(xref.clone_ref(py));
@@ -233,80 +225,60 @@ impl IntoPy<fastobo::ast::XrefList> for XrefList {
 }
 
 impl IntoPy<fastobo::ast::XrefList> for &XrefList {
-    fn into_py(self, py: Python) -> fastobo::ast::XrefList {
+    fn into_py<'py>(self, py: Python) -> fastobo::ast::XrefList {
         self.xrefs
             .iter()
-            .map(|xref| xref.as_ref(py).borrow().clone_py(py).into_py(py))
+            .map(|xref| xref.bind(py).borrow().clone_py(py).into_py(py))
             .collect()
     }
 }
 
-impl ToPyObject for XrefList {
-    fn to_object(&self, py: Python) -> PyObject {
-        let list = self.xrefs
-            .iter()
-            .map(|xref| xref.clone_py(py))
-            .collect();
-        IntoPy::into_py(XrefList::new(list), py)
-    }
-}
-
 #[listlike(field = "xrefs", type = "Py<Xref>")]
 #[pymethods]
 impl XrefList {
     #[new]
-    fn __init__(xrefs: Option<&PyAny>) -> PyResult<Self> {
+    #[pyo3(signature = (xrefs = None))]
+    fn __init__<'py>(xrefs: Option<&Bound<'py, PyAny>>) -> PyResult<Self> {
         if let Some(x) = xrefs {
-            let gil = Python::acquire_gil();
-            Self::collect(gil.python(), x)
+            Python::with_gil(|py| Self::collect(py, x))
         } else {
             Ok(Self::new(Vec::new()))
         }
     }
-}
 
-#[pyproto]
-impl PyObjectProtocol for XrefList {
-    fn __repr__(&self) -> PyResult<PyObject> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        if self.xrefs.is_empty() {
-            Ok("XrefList()".to_object(py))
+    fn __repr__(slf: PyRef<Self>) -> PyResult<Bound<PyAny>> {
+        if slf.xrefs.is_empty() {
+            Ok(PyString::intern(slf.py(), "XrefList()").into_any())
         } else {
-            let fmt = PyString::new(py, "XrefList({!r})").to_object(py);
-            fmt.call_method1(py, "format", (&self.xrefs.to_object(py),))
+            let fmt = PyString::intern(slf.py(), "XrefList({!r})");
+            fmt.call_method1("format", (&slf.xrefs,))
         }
-
     }
 
     fn __str__(&self) -> PyResult<String> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let frame: fastobo::ast::XrefList = self.clone_py(py).into_py(py);
+        let frame: fastobo::ast::XrefList = Python::with_gil(|py| self.clone_py(py).into_py(py));
         Ok(frame.to_string())
     }
-}
 
-#[pyproto]
-impl PySequenceProtocol for XrefList {
     fn __len__(&self) -> PyResult<usize> {
         Ok(self.xrefs.len())
     }
 
     fn __getitem__(&self, index: isize) -> PyResult<Py<Xref>> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
         if index < self.xrefs.len() as isize {
-            Ok(self.xrefs[index as usize].clone_ref(py))
+            Python::with_gil(|py| Ok(self.xrefs[index as usize].clone_ref(py)))
         } else {
             Err(PyIndexError::new_err("list index out of range"))
         }
     }
 
-    fn __contains__(&self, item: &PyAny) -> PyResult<bool> {
+    fn __contains__<'py>(&self, item: &Bound<'py, PyAny>) -> PyResult<bool> {
         if let Ok(xref) = item.extract::<Py<Xref>>() {
             let py = item.py();
-            Ok(self.xrefs.iter().any(|x| *x.as_ref(py).borrow() == *xref.as_ref(py).borrow()))
+            Ok(self
+                .xrefs
+                .iter()
+                .any(|x| (*x.bind(py).borrow()).eq_py(&xref.bind(py).borrow(), py)))
         } else {
             let ty = item.get_type().name()?;
             let msg = format!("'in <XrefList>' requires Xref as left operand, not {}", ty);
