@@ -8,25 +8,21 @@ use std::sync::Mutex;
 
 use pyo3::exceptions::PyOSError;
 use pyo3::exceptions::PyTypeError;
-use pyo3::gc::PyGCProtocol;
 use pyo3::gc::PyTraverseError;
 use pyo3::gc::PyVisit;
 use pyo3::prelude::*;
 use pyo3::types::PyBytes;
-use pyo3::AsPyPointer;
-use pyo3::PyDowncastError;
 use pyo3::PyObject;
-use pyo3::PyNativeType;
 
 // ---------------------------------------------------------------------------
 
 #[macro_export]
 macro_rules! transmute_file_error {
-    ($self:ident, $e:ident, $msg:expr, $py:expr) => ({
+    ($self:ident, $e:ident, $msg:expr, $py:expr) => {{
         // Attempt to transmute the Python OSError to an actual
         // Rust `std::io::Error` using `from_raw_os_error`.
-        if $e.is_instance::<PyOSError>($py) {
-            if let Ok(code) = &$e.pvalue($py).getattr("errno") {
+        if $e.is_instance_of::<PyOSError>($py) {
+            if let Ok(code) = &$e.value($py).getattr("errno") {
                 if let Ok(n) = code.extract::<i32>() {
                     return Err(IoError::from_raw_os_error(n));
                 }
@@ -38,24 +34,27 @@ macro_rules! transmute_file_error {
         // generic Rust error instead.
         $e.restore($py);
         Err(IoError::new(std::io::ErrorKind::Other, $msg))
-    });
+    }};
 }
 
 // ---------------------------------------------------------------------------
 
 /// A wrapper around a readable Python file borrowed within a GIL lifetime.
-pub struct PyFileRead<'p> {
-    file: &'p PyAny,
+pub struct PyFileRead<'py> {
+    file: Bound<'py, PyAny>,
 }
 
-impl<'p> PyFileRead<'p> {
-    pub fn from_ref(file: &'p PyAny) -> PyResult<PyFileRead<'p>> {
+impl<'py> PyFileRead<'py> {
+    pub fn from_ref(file: &Bound<'py, PyAny>) -> PyResult<PyFileRead<'py>> {
         let res = file.call_method1("read", (0,))?;
-        if res.cast_as::<PyBytes>().is_ok() {
-            Ok(PyFileRead { file })
+        if res.downcast::<PyBytes>().is_ok() {
+            Ok(PyFileRead { file: file.clone() })
         } else {
             let ty = res.get_type().name()?.to_string();
-            Err(PyTypeError::new_err(format!("expected bytes, found {}", ty)))
+            Err(PyTypeError::new_err(format!(
+                "expected bytes, found {}",
+                ty
+            )))
         }
     }
 }
@@ -65,7 +64,7 @@ impl<'p> Read for PyFileRead<'p> {
         match self.file.call_method1("read", (buf.len(),)) {
             Ok(obj) => {
                 // Check `fh.read` returned bytes, else raise a `TypeError`.
-                if let Ok(bytes) = obj.extract::<&PyBytes>() {
+                if let Ok(bytes) = obj.downcast::<PyBytes>() {
                     let b = bytes.as_bytes();
                     (&mut buf[..b.len()]).copy_from_slice(b);
                     Ok(b.len())
@@ -80,9 +79,7 @@ impl<'p> Read for PyFileRead<'p> {
                 }
             }
             Err(e) => {
-                transmute_file_error!(
-                    self, e, "read method failed", self.file.py()
-                )
+                transmute_file_error!(self, e, "read method failed", self.file.py())
             }
         }
     }
@@ -91,24 +88,24 @@ impl<'p> Read for PyFileRead<'p> {
 // ---------------------------------------------------------------------------
 
 /// A wrapper around a writable Python file borrowed within a GIL lifetime.
-pub struct PyFileWrite<'p> {
-    file: &'p PyAny,
+pub struct PyFileWrite<'py> {
+    file: Bound<'py, PyAny>,
 }
 
-impl<'p> PyFileWrite<'p> {
-    pub fn from_ref(file: &'p PyAny) -> PyResult<PyFileWrite<'p>> {
+impl<'py> PyFileWrite<'py> {
+    pub fn from_ref(file: &Bound<'py, PyAny>) -> PyResult<PyFileWrite<'py>> {
         file.call_method1("write", (PyBytes::new(file.py(), b""),))
-            .map(|_| PyFileWrite { file })
+            .map(|_| PyFileWrite { file: file.clone() })
     }
 }
 
-impl<'p> Write for PyFileWrite<'p> {
+impl<'py> Write for PyFileWrite<'py> {
     fn write(&mut self, buf: &[u8]) -> Result<usize, IoError> {
         let bytes = PyBytes::new(self.file.py(), buf);
         match self.file.call_method1("write", (bytes,)) {
             Ok(obj) => {
                 // Check `fh.write` returned int, else raise a `TypeError`.
-                if let Ok(len) = usize::extract(&obj) {
+                if let Ok(len) = obj.extract::<usize>() {
                     Ok(len)
                 } else {
                     let ty = obj.get_type().name()?.to_string();
@@ -121,9 +118,7 @@ impl<'p> Write for PyFileWrite<'p> {
                 }
             }
             Err(e) => {
-                transmute_file_error!(
-                    self, e, "write method failed", self.file.py()
-                )
+                transmute_file_error!(self, e, "write method failed", self.file.py())
             }
         }
     }
@@ -132,9 +127,7 @@ impl<'p> Write for PyFileWrite<'p> {
         match self.file.call_method0("flush") {
             Ok(_) => Ok(()),
             Err(e) => {
-                transmute_file_error!(
-                    self, e, "flush method failed", self.file.py()
-                )
+                transmute_file_error!(self, e, "flush method failed", self.file.py())
             }
         }
     }
@@ -148,15 +141,18 @@ pub struct PyFileGILRead {
 }
 
 impl PyFileGILRead {
-    pub fn from_ref(file: &PyAny) -> PyResult<PyFileGILRead> {
+    pub fn from_ref<'py>(file: &Bound<'py, PyAny>) -> PyResult<PyFileGILRead> {
         let res = file.call_method1("read", (0,))?;
-        if res.cast_as::<PyBytes>().is_ok() {
+        if res.downcast::<PyBytes>().is_ok() {
             Ok(PyFileGILRead {
-                file: Mutex::new(file.to_object(file.py())),
+                file: Mutex::new(file.clone().unbind()),
             })
         } else {
             let ty = res.get_type().name()?.to_string();
-            Err(PyTypeError::new_err(format!("expected bytes, found {}", ty)))
+            Err(PyTypeError::new_err(format!(
+                "expected bytes, found {}",
+                ty
+            )))
         }
     }
 
@@ -167,27 +163,28 @@ impl PyFileGILRead {
 
 impl Read for PyFileGILRead {
     fn read(&mut self, buf: &mut [u8]) -> Result<usize, IoError> {
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let file = self.file.lock().unwrap();
-        match file.to_object(py).call_method1(py, "read", (buf.len(),)) {
-            Ok(obj) => {
-                // Check `fh.read` returned bytes, else raise a `TypeError`.
-                if let Ok(bytes) = obj.cast_as::<PyBytes>(py) {
-                    let b = bytes.as_bytes();
-                    (&mut buf[..b.len()]).copy_from_slice(b);
-                    Ok(b.len())
-                } else {
-                    let ty = obj.as_ref(py).get_type().name()?.to_string();
-                    let msg = format!("expected bytes, found {}", ty);
-                    PyTypeError::new_err(msg).restore(py);
-                    Err(IoError::new(
-                        std::io::ErrorKind::Other,
-                        "fh.read did not return bytes",
-                    ))
+        Python::with_gil(|py| {
+            let guard = self.file.lock().unwrap();
+            let file = guard.bind(py);
+            match file.call_method1("read", (buf.len(),)) {
+                Ok(obj) => {
+                    // Check `fh.read` returned bytes, else raise a `TypeError`.
+                    if let Ok(bytes) = obj.downcast::<PyBytes>() {
+                        let b = bytes.as_bytes();
+                        (&mut buf[..b.len()]).copy_from_slice(b);
+                        Ok(b.len())
+                    } else {
+                        let ty = obj.get_type().name()?.to_string();
+                        let msg = format!("expected bytes, found {}", ty);
+                        PyTypeError::new_err(msg).restore(py);
+                        Err(IoError::new(
+                            std::io::ErrorKind::Other,
+                            "fh.read did not return bytes",
+                        ))
+                    }
                 }
+                Err(e) => transmute_file_error!(self, e, "read method failed", py),
             }
-            Err(e) => transmute_file_error!(self, e, "read method failed", py)
-        }
+        })
     }
 }
