@@ -38,6 +38,16 @@ def _load_tests_from_module(tests, module, globs, setUp=None, tearDown=None):
 def load_tests(loader, tests, ignore):
     """load_test function used by unittest to find the doctests"""
 
+    _current_cwd = os.getcwd()
+
+    def setUp(self):
+        warnings.simplefilter("ignore")
+        os.chdir(os.path.realpath(os.path.join(__file__, "..", "data")))
+
+    def tearDown(self):
+        os.chdir(_current_cwd)
+        warnings.simplefilter(warnings.defaultaction)
+
     globs = {
         "fastobo": fastobo,
         "datetime": datetime,
@@ -49,7 +59,7 @@ def load_tests(loader, tests, ignore):
     }
 
     if not sys.argv[0].endswith('green'):
-        tests = _load_tests_from_module(tests, fastobo, globs)
+        tests = _load_tests_from_module(tests, fastobo, globs, setUp, tearDown)
     return tests
 
 
