@@ -1,7 +1,9 @@
 # coding: utf-8
 
+import io
 import os
 import unittest
+import pathlib
 
 import fastobo
 
@@ -61,6 +63,10 @@ class TestLoad(unittest.TestCase):
     def test_threading_invalid(self):
         self.assertRaises(ValueError, fastobo.load, MS, threads=-1)
 
+    def test_load_path(self):
+        doc = fastobo.load(pathlib.Path(MS))
+        self.assertEqual(len(doc), MS_FRAMES)
+
 
 class TestIter(unittest.TestCase):
 
@@ -74,6 +80,10 @@ class TestIter(unittest.TestCase):
         with open(MS) as f:
             self.assertRaises(TypeError, fastobo.iter, f)
 
+    def test_syntax_error(self):
+        f = io.BytesIO(b"format-version: 1.4\ndate: 05:20:2021 12:00\n")
+        self.assertRaises(SyntaxError, fastobo.iter, f)
+
     def test_threading_single(self):
         frame_count = sum(1 for _ in fastobo.iter(MS, threads=1))
         self.assertEqual(frame_count, MS_FRAMES)
@@ -101,6 +111,11 @@ class TestIter(unittest.TestCase):
     def test_threading_invalid(self):
         self.assertRaises(ValueError, fastobo.iter, MS, threads=-1)
 
+    def test_iter_path(self):
+        n = sum(1 for _ in fastobo.iter(pathlib.Path(MS)))
+        self.assertEqual(n, MS_FRAMES)
+
+
 class TestLoads(unittest.TestCase):
 
     @classmethod
