@@ -37,6 +37,17 @@ class TestXref(unittest.TestCase):
             'ISBN:0321842685 "Hacker\'s Delight (2nd Edition)"'
         )
 
+    def test_eq(self):
+        i1 = fastobo.id.UnprefixedIdent('a')
+        i2 = fastobo.id.UnprefixedIdent('b')
+        x1 = self.type(i1)
+        self.assertEqual(x1, x1)
+        x2 = self.type(i1)
+        self.assertIsNot(x1, x2)
+        self.assertEqual(x1, x2)
+        x3 = self.type(i2)
+        self.assertNotEqual(x1, x2)
+
 
 class TestXrefList(unittest.TestCase):
 
@@ -73,6 +84,18 @@ class TestXrefList(unittest.TestCase):
         self.assertEqual(str(self.type([x1])), '[{}]'.format(x1))
         self.assertEqual(str(self.type([x1, x2])), '[{}, {}]'.format(x1, x2))
 
+    def test_append(self):
+        x1, x2 = self.x1, self.x2
+        l = self.type()
+        self.assertEqual(len(l), 0)
+        l.append(x1)
+        self.assertEqual(len(l), 1)
+        self.assertEqual(l[0], x1)
+        l.append(x2)
+        self.assertEqual(len(l), 2)
+        self.assertEqual(l[0], x1)
+        self.assertEqual(l[1], x2)
+
     def test_contains(self):
         x1, x2 = self.x1, self.x2
         l1 = self.type()
